"""The Tag problem. Implemented according to the paper `Anytime Point-Based
Approximations for Large POMDPs <https://arxiv.org/pdf/1110.0027.pdf>`_.

Transition model: the robot moves deterministically. The target's movement
    depends on the robot; With Pr=0.8 the target moves away from the robot,
    and with Pr=0.2, the target stays at the same place. The target never
    moves closer to the robot.
"""
import copy
import pomdp_py
import pomdp_problems.tag.constants as constants
from pomdp_problems.tag.domain.action import *

class TagTransitionModel(pomdp_py.TransitionModel):

    def __init__(self,
                 grid_map,
                 target_motion_policy):
        self._grid_map = grid_map
        self.target_motion_policy = target_motion_policy

    @classmethod
    def if_move_by(cls, grid_map, position, action):
        if isinstance(action, MotionAction):
            dx, dy = action.motion
            next_position = (position[0] + dx,
                             position[1] + dy)
            if grid_map.valid_pose(next_position):
                return next_position
        return position

    def probability(self, next_state, state, action, **kwargs):
        # Robot motion
        expected_robot_position = TagTransitionModel.if_move_by(self._grid_map,
                                                                state.robot_position,
                                                                action)
        if expected_robot_position != next_state.robot_position:
            return constants.EPSILON

        if isinstance(action, TagAction):
            if next_state.target_position == next_state.robot_position:
                if next_state.target_found:
                    return 1.0 - constants.EPSILON
                else:
                    return constants.EPSILON
            else:
                if next_state.target_found:
                    return constants.EPSILON
                else:
                    return 1.0 - constants.EPSILON

        # Target motion
        valid_target_motion_actions = self._grid_map.valid_motions(state.target_position)
        return self.target_motion_policy.probability(next_state.target_position,
                                                     state.target_position,
                                                     state.robot_position,
                                                     valid_target_motion_actions)

    def sample(self, state, action, argmax=False):
        # Robot motion
        next_state = copy.deepcopy(state)
        next_state.robot_position = TagTransitionModel.if_move_by(self._grid_map,
                                                                  state.robot_position,
                                                                  action)

        # If Tag action
        if isinstance(action, TagAction):
            if not state.target_found:
                if state.robot_position == state.target_position:
                    next_state.target_found = True
            return next_state

        # Target motion
        valid_target_motion_actions = self._grid_map.valid_motions(state.target_position)
        if not argmax:
            next_state.target_position = self.target_motion_policy.random(state.robot_position,
                                                                          state.target_position,
                                                                          valid_target_motion_actions)
        else:
            next_state.target_position = self.target_motion_policy.mpe(state.robot_position,
                                                                       state.target_position,
                                                                       valid_target_motion_actions)
        return next_state

    def argmax(self, state, action, **kwargs):
        return self.sample(state, action, argmax=True)



"""Optional grid map to assist collision avoidance during planning."""

from pomdp_problems.multi_object_search.models.transition_model import RobotTransitionModel
from pomdp_problems.multi_object_search.domain.action import *
from pomdp_problems.multi_object_search.domain.state import *

class GridMap:
    """This map assists the agent to avoid planning invalid
    actions that will run into obstacles. Used if we assume
    the agent has a map. This map does not contain information
    about the object locations."""
    def __init__(self, width, length, obstacles):
        """
        Args:
            obstacles (dict): Map from objid to (x,y); The object is
                                   supposed to be an obstacle.
            width (int): width of the grid map
            length (int): length of the grid map 
        """
        self.width = width
        self.length = length
        self._obstacles = obstacles
        # An MosOOState that only contains poses for obstacles;
        # This is to allow calling RobotTransitionModel.if_move_by
        # function.
        self._obstacle_states = {
            objid: ObjectState(objid, "obstacle", self._obstacles[objid])
            for objid in self._obstacles
        }
        # set of obstacle poses
        self.obstacle_poses = set({self._obstacles[objid]
                                   for objid in self._obstacles})        

    def valid_motions(self, robot_id, robot_pose, all_motion_actions):
        """
        Returns a set of MotionAction(s) that are valid to
        be executed from robot pose (i.e. they will not bump
        into obstacles). The validity is determined under
        the assumption that the robot dynamics is deterministic.
        """
        state = MosOOState(self._obstacle_states)
        state.set_object_state(robot_id,
                               RobotState(robot_id, robot_pose, None, None))

        valid = set({})
        for motion_action in all_motion_actions:
            if not isinstance(motion_action, MotionAction):
                raise ValueError("This (%s) is not a motion action" % str(motion_action))

            next_pose = RobotTransitionModel.if_move_by(robot_id, state,
                                                        motion_action, (self.width, self.length))
            if next_pose != robot_pose:
                # robot moved --> valid motion
                valid.add(motion_action)
        return valid



"""This module describes components of the
OO-POMDP interface in `pomdp_py`.

An OO-POMDP is a specific type of POMDP where
the state and observation spaces are factored
by objects. As a result, the transition, observation,
and belief distributions are all factored by objects.
A main benefit of using OO-POMDP is that the
object factoring reduces the scaling of belief
space from exponential to linear as the number
of objects increases. See :cite:`wandzel2019multi`."""

from pomdp_py.framework.basics cimport POMDP, State, Action, Observation,\
    ObservationModel, TransitionModel, GenerativeDistribution
from pomdp_py.utils.cython_utils cimport det_dict_hash
import collections
import copy

cdef class OOPOMDP(POMDP):
    """
    An OO-POMDP is again defined by an agent and an environment.

    __init__(self, agent, env, name="OOPOMDP")
    """
    def __init__(self, agent, env, name="OOPOMDP"):
        super().__init__(agent, env, name=name)

cdef class ObjectState(State):
    """
    This is the result of OOState factoring; A state
    in an OO-POMDP is made up of ObjectState(s), each with
    an `object class` (str) and a set of `attributes` (dict).
    """
    def __init__(self, objclass, attributes):
        """
        class: "class",
        attributes: {
            "attr1": value,  # value should be hashable
            ...
        }
        """
        self.objclass = objclass
        self.attributes = attributes
        self._hashcache = -1

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return '%s::(%s,%s)' % (str(self.__class__.__name__),
                                str(self.objclass),
                                str(self.attributes))

    def __hash__(self):
        """__hash__(self)
        Hash the ObjectState.
        For more efficient hashing for your specific domain, please overwrite this function."""
        if self._hashcache == -1:
            self._hashcache = det_dict_hash(self.attributes)
        return self._hashcache

    def __eq__(self, other):
        if not isinstance(other, ObjectState):
            return False
        return self.objclass == other.objclass\
            and self.attributes == other.attributes

    def __getitem__(self, attr):
        """__getitem__(self, attr)
        Returns the attribute"""
        return self.attributes[attr]

    def __setitem__(self, attr, value):
        """__setitem__(self, attr, value)
        Sets the attribute `attr` to the given value."""
        self.attributes[attr] = value

    def __len__(self):
        return len(self.attributes)

    def copy(self):
        """copy(self)
        Copies this ObjectState."""
        return ObjectState(self.objclass, copy.deepcopy(self.attributes))


cdef class OOState(State):

    """
    State that can be factored by objects, that is, to ObjectState(s).

    Note: to change the state of an object, you can use set_object_state.  Do
    not directly assign the object state by e.g. oostate.object_states[objid] =
    object_state, because it will cause the hashcode to be incorrect in the
    oostate after the change.

    __init__(self, object_states)
    """

    def __init__(self, object_states):
        """
        objects_states: dictionary of dictionaries; Each dictionary represents an object state:
            { ID: ObjectState }
        """
        # internally, objects are sorted by ID.
        self.object_states = object_states
        self._hashcache = -1

    @property
    def situation(self):
        """situation(selppf)
        This is a `frozenset` which can be used to identify
        the situation of this state since it supports hashing."""
        return frozenset(self.object_states.items())

    def __str__(self):
        return '%s::[%s]' % (str(self.__class__.__name__),
                             str(self.object_states))

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return isinstance(other, OOState)\
            and self.object_states == other.object_states

    def __hash__(self):
        """__hash__(self)
        Hash the ObjectState.
        For more efficient hashing for your specific domain, please overwrite this function."""
        if self._hashcache == -1:
            self._hashcache = det_dict_hash(self.object_states)
        return self._hashcache

    def __getitem__(self, index):
        raise NotImplemented

    def __len__(self):
        raise NotImplemented

    def get_object_state(self, objid):
        """get_object_state(self, objid)
        Returns the ObjectState for given object."""
        return self.object_states[objid]

    def set_object_state(self, objid, object_state):
        """set_object_state(self, objid, object_state)
        Sets the state of the given
        object to be the given object state (ObjectState)
        """
        # Deprecation Warning: set_object_state is not valid because OOState is not mutable.
        self.object_states[objid] = object_state
        self._hashcache = -1

    def get_object_class(self, objid):
        """get_object_class(self, objid)
        Returns the class of requested object"""
        return self.object_states[objid].objclass

    def get_object_attribute(self, objid, attr):
        """get_object_attribute(self, objid, attr)
        Returns the attributes of requested object"""
        return self.object_states[objid][attr]

    def copy(self):
        """copy(self)
        Copies the state."""
        return OOState(copy.deepcopy(self.object_states))

    def s(self, objid):
        """convenient alias function"""
        return self.object_states[objid]

cdef class OOTransitionModel(TransitionModel):

    """
    :math:`T(s' | s, a) = \prod_i T(s_i' | s, a)`

    __init__(self, transition_models):
    Args:
        transition_models (dict) objid -> transition_model
    """

    def __init__(self, transition_models):
        """
        transition_models (dict) objid -> transition_model
        """
        self._transition_models = transition_models

    def probability(self, next_state, state, action, **kwargs):
        """probability(self, next_state, state, action, **kwargs)
        Returns :math:`T(s' | s, a)
        """
        if not isinstance(next_state, OOState):
            raise ValueError("next_state must be OOState")
        if not isinstance(state, OOState):
            raise ValueError("state must be OOState")
        if next_state.object_states.keys() != state.object_states.keys():
            raise ValueError("object types modified between states")
        trans_prob = 1.0
        for objid in self._transition_models:
            object_state = state.object_states[objid]
            next_object_state = next_state.object_states[objid]
            trans_prob = trans_prob * self._transition_models[objid].probability(next_object_state, state, action, **kwargs)
        return trans_prob

    def sample(self, state, action, argmax=False, **kwargs):
        """
        sample(self, state, action, argmax=False, **kwargs)
        Returns random next_state"""
        if not isinstance(state, OOState):
            raise ValueError("state must be OOState")
        object_states = {}
        for objid in state.object_states:
            if objid not in self._transition_models:
                # no transition model provided for this object. Then no transition happens.
                object_states[objid] = copy.deepcopy(state.object_states[objid])
                continue
            if argmax:
                next_object_state = self._transition_models[objid].argmax(state, action, **kwargs)
            else:
                next_object_state = self._transition_models[objid].sample(state, action, **kwargs)
            object_states[objid] = next_object_state
        return OOState(object_states)

    def argmax(self, state, action, **kwargs):
        """
        argmax(self, state, action, **kwargs)
        Returns the most likely next state"""
        return self.sample(state, action, argmax=True, **kwargs)

    def __getitem__(self, objid):
        """__getitem__(self, objid)
        Returns transition model for given object"""
        return self._transition_models[objid]

    @property
    def transition_models(self):
        """transition_models(self)"""
        return self._transition_models

cdef class OOObservation(Observation):
    def factor(self, next_state, action, **kwargs):
        """factor(self, next_state, action, **kwargs)
        Factors the observation by objects.
        That is, :math:`z\mapsto z_1,\cdots,z_n`

        Args:
            next_state (OOState): given state
            action (Action): given action
        Returns:
            dict: map from object id to a `pomdp_py.Observation`.
        """
        raise NotImplemented
    @classmethod
    def merge(cls, object_observations, next_state, action, **kwargs):
        """merge(cls, object_observations, next_state, action, **kwargs)
        Merges the factored `object_observations` into a
        single OOObservation.

        Args:
            object_observations (dict): map from object id to a `pomdp_py.Observation`.
            next_state (OOState): given state
            action (Action): given action
        Returns:
            OOObservation: the merged observation.
        """
        raise NotImplemented

cdef class OOObservationModel(ObservationModel):

    """
    :math:`O(z | s', a) = \prod_i O(z_i' | s', a)`

    __init__(self, observation_models):
    Args:
        observation_models (dict) objid -> observation_model
    """

    def __init__(self, observation_models, merge_func=None):#, factor_observation_func, merge_observations_func):
        """
        observation_models (dict) objid -> observation_model
        factor_observation_func: (observation, objid, next_state, action) -> {objid->observations}
        merge_observations_func: (factored_observations, next_state, action) -> observation
        """
        self._observation_models = observation_models
        self._merge_func = merge_func


    def probability(self, observation, next_state, action, **kwargs):
        """
        probability(self, observation, next_state, action, **kwargs)
        Returns :math:`O(z | s', a)`.
        """
        if not isinstance(next_state, OOState):
            raise ValueError("state must be OOState")
        obsrv_prob = 1.0
        factored_observations = observation.factor(next_state, action) #self._factor_observation_func(observation, next_state)
        for objid in next_state.object_states:
            obsrv_prob = obsrv_prob * self._observation_models[objid].probability(factored_observations[objid],
                                                                                  next_state, action, **kwargs)
        return obsrv_prob

    def sample(self, next_state, action, argmax=False, **kwargs):
        """
        sample(self, next_state, action, argmax=False, **kwargs)
        Returns random observation"""
        if not isinstance(next_state, OOState):
            raise ValueError("state must be OOState")
        factored_observations = {}
        for objid in self._observation_models:
            if not argmax:
                observation = self._observation_models[objid].sample(next_state,
                                                                     action, **kwargs)
            else:
                observation = self._observation_models[objid].argmax(next_state,
                                                                     action, **kwargs)
            factored_observations[objid] = observation
        if self._merge_func is None:
            return factored_observations
        else:
            return self._merge_func(factored_observations, next_state, action, **kwargs)

    def argmax(self, next_state, action, **kwargs):
        """
        argmax(self, next_state, action, **kwargs)
        Returns most likely observation"""
        return self.sample(next_state, action, argmax=True, **kwargs)

    def __getitem__(self, objid):
        """__getitem__(self, objid)
        Returns observation model for given object"""
        return self._observation_models[objid]

    @property
    def observation_models(self):
        """observation_models(self)"""
        return self._observation_models


cdef class OOBelief(GenerativeDistribution):
    """
    Belief factored by objects.
    """
    def __init__(self, object_beliefs):
        """
        object_beliefs (objid -> GenerativeDistribution)
        """
        self._object_beliefs = object_beliefs

    def __getitem__(self, state):
        """__getitem__(self, state)
        Returns belief probability of given state"""
        if not isinstance(state, OOState):
            raise ValueError("state must be OOState")
        belief_prob = 1.0
        for objid in self._object_beliefs:
            object_state = state.object_states[objid]
            belief_prob = belief_prob * self._object_beliefs[objid].probability(object_state)
        return belief_prob

    def random(self, return_oostate=True, **kwargs):
        """random(self, return_oostate=False, **kwargs)
        Returns a random state"""
        object_states = {}
        for objid in self._object_beliefs:
            object_states[objid] = self._object_beliefs[objid].random(**kwargs)
        if return_oostate:
            return OOState(object_states)
        else:
            return object_states

    def mpe(self, return_oostate=True, **kwargs):
        """mpe(self, return_oostate=False, **kwargs)
        Returns most likely state."""
        object_states = {}
        for objid in self._object_beliefs:
            object_states[objid] = self._object_beliefs[objid].mpe(**kwargs)
        if return_oostate:
            return OOState(object_states)
        else:
            return object_states

    def __setitem__(self, oostate, value):
        """__setitem__(self, oostate, value)
        Sets the probability of a given `oostate` to `value`.
        Note always feasible."""
        raise NotImplemented

    def object_belief(self, objid):
        """object_belief(self, objid)
        Returns the belief (GenerativeDistribution) for the given object."""
        return self._object_beliefs[objid]

    def set_object_belief(self, objid, belief):
        """set_object_belief(self, objid, belief)
        Sets the belief of object to be the given `belief` (GenerativeDistribution)"""
        self._object_beliefs[objid] = belief

    @property
    def object_beliefs(self):
        """object_beliefs(self)"""
        return self._object_beliefs

    def b(self, objid):
        """convenient alias function call"""
        return self._object_beliefs[objid]



import pomdp_py
import random
from pomdp_problems.tag.env.env import *
from pomdp_problems.tag.env.visual import *
from pomdp_problems.tag.agent.agent import *
from pomdp_problems.tag.example_worlds import *
import time

class TagProblem(pomdp_py.POMDP):

    def __init__(self,
                 init_robot_position,
                 init_target_position,
                 grid_map,
                 pr_stay=0.2,
                 small=1,
                 big=10,
                 prior="uniform",
                 belief_type="hist",
                 num_particles=6):
        init_state = TagState(init_robot_position,
                              init_target_position,
                              False)
        env = TagEnvironment(init_state,
                             grid_map,
                             pr_stay=pr_stay,
                             small=1,
                             big=10)
        if prior == "uniform":
            prior = {}
        elif prior == "informed":
            prior = {init_target_position: 1.0}
        else:
            raise ValueError("Unrecognized prior type: %s" % prior)

        if belief_type == "particles":
            init_belief = initialize_particles_belief(grid_map, init_robot_position,
                                                      prior=prior, num_particles=num_particles)
        else:
            init_belief = initialize_belief(grid_map, init_robot_position, prior=prior)
        agent = TagAgent(init_belief,
                         grid_map,
                         pr_stay=pr_stay,
                         small=1,
                         big=10)
        super().__init__(agent, env,
                         name="TagProblem")


def solve(problem,
          planner_type="pouct",
          max_depth=10,  # planning horizon
          discount_factor=0.99,
          planning_time=1.,       # amount of time (s) to plan each step
          exploration_const=1000, # exploration constant
          visualize=True,
          max_time=120,  # maximum amount of time allowed to solve the problem
          max_steps=500):  # maximum number of planning steps the agent can take.
    if planner_type == "pouct":
        planner = pomdp_py.POUCT(max_depth=max_depth,
                                 discount_factor=discount_factor,
                                 planning_time=planning_time,
                                 exploration_const=exploration_const,
                                 rollout_policy=problem.agent.policy_model)
    else:
        planner = pomdp_py.POMCP(max_depth=max_depth,
                                 discount_factor=discount_factor,
                                 planning_time=planning_time,
                                 exploration_const=exploration_const,
                                 rollout_policy=problem.agent.policy_model)
    if visualize:
        viz = TagViz(problem.env, controllable=False)
        if viz.on_init() == False:
            raise Exception("Environment failed to initialize")
        viz.update(None,
                   None,
                   problem.agent.cur_belief)
        viz.on_render()

    _discount = 1.0
    _time_used = 0
    _find_actions_count = 0
    _total_reward = 0  # total, undiscounted reward
    _total_discounted_reward = 0
    for i in range(max_steps):
        # Plan action
        _start = time.time()
        real_action = planner.plan(problem.agent)
        _time_used += time.time() - _start
        if _time_used > max_time:
            break  # no more time to update.

        # Execute action
        reward = problem.env.state_transition(real_action, execute=True)

        # Receive observation
        _start = time.time()
        real_observation = \
            problem.env.provide_observation(problem.agent.observation_model, real_action)

        # Updates
        problem.agent.clear_history()  # truncate history
        problem.agent.update_history(real_action, real_observation)
        planner.update(problem.agent, real_action, real_observation)
        if planner_type == "pouct":
            belief_update(problem.agent, real_action, real_observation)
        _time_used += time.time() - _start

        # Info and render
        _total_reward += reward
        _total_discounted_reward += reward * _discount
        _discount = _discount * discount_factor
        print("==== Step %d ====" % (i+1))
        print("Action: %s" % str(real_action))
        print("Observation: %s" % str(real_observation))
        print("Reward: %s" % str(reward))
        print("Reward (Cumulative): %s" % str(_total_reward))
        print("Reward (Discounted): %s" % str(_total_discounted_reward))
        print("Find Actions Count: %d" %  _find_actions_count)
        if isinstance(planner, pomdp_py.POUCT):
            print("__num_sims__: %d" % planner.last_num_sims)

        if visualize:
            viz.update(real_action,
                       real_observation,
                       problem.agent.cur_belief)
            viz.on_loop()
            viz.on_render()

        # Termination check
        if problem.env.state.target_found:
            print("Done!")
            break
        if _time_used > max_time:
            print("Maximum time reached.")
            break
        if _discount*10 < 1e-4:
            print("Discount factor already too small")
            break

    return _total_discounted_reward


def main():
    worldstr, robotstr = world0
    grid_map = GridMap.from_str(worldstr)
    free_cells = grid_map.free_cells()
    init_robot_position = random.sample(free_cells, 1)[0]
    init_target_position = random.sample(free_cells, 1)[0]

    problem = TagProblem(init_robot_position,
                         init_target_position,
                         grid_map,
                         pr_stay=0.2,
                         small=1,
                         big=10,
                         prior="uniform",
                         belief_type="histogram")
    solve(problem,
          max_depth=15,
          discount_factor=0.95,
          planning_time=0.8,
          exploration_const=20,
          visualize=True,
          max_time=360,
          max_steps=251,
          planner_type="pouct")


if __name__ == "__main__":
    main()



"""
RockSample(n,k) problem

Origin: Heuristic Search Value Iteration for POMDPs (UAI 2004)

Description:

State space:

    Position {(1,1),(1,2),...(n,n)}
    :math:`\\times` RockType_1 :math:`\\times` RockType_2, ..., :math:`\\times` RockType_k
    where RockType_i = {Good, Bad}
    :math:`\\times` TerminalState

    (basically, the positions of rocks are known to the robot,
     but not represented explicitly in the state space. Check_i
     will smartly check the rock i at its location.)

Action space:

    North, South, East, West, Sample, Check_1, ..., Check_k
    The first four moves the agent deterministically
    Sample: samples the rock at agent's current location
    Check_i: receives a noisy observation about RockType_i
    (noise determined by eta (:math:`\eta`). eta=1 -> perfect sensor; eta=0 -> uniform)

Observation: observes the property of rock i when taking Check_i.

Reward: +10 for Sample a good rock. -10 for Sampling a bad rock.
        Move to exit area +10. Other actions have no cost or reward.

Initial belief: every rock has equal probability of being Good or Bad.
"""
from pomdp_py.framework.basics cimport *
from pomdp_py.algorithms.po_uct cimport *
from pomdp_py.representations.belief.particles cimport *
import pomdp_py
import random
import math
import numpy as np
import sys
import copy

EPSILON = 1e-9

def euclidean_dist(p1, p2):
    return math.sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)

cdef class RockType:
    GOOD = 'good'
    BAD = 'bad'
    @staticmethod
    def invert(rocktype):
        if rocktype == 'good':
            return 'bad'
        else:
            return 'good'
        # return 1 - rocktype
    @staticmethod
    def random(p=0.5):
        if random.uniform(0,1) >= p:
            return RockType.GOOD
        else:
            return RockType.BAD

cdef class RSState(State):
    cdef public tuple position
    cdef public tuple rocktypes
    cdef public bint terminal
    def __init__(self, position, rocktypes, terminal=False):
        """
        position (tuple): (x,y) position of the rover on the grid.
        rocktypes: tuple of size k. Each is either Good or Bad.
        terminal (bool): The robot is at the terminal state.

        (It is so true that the agent's state doesn't need to involve the map!)

        x axis is horizontal. y axis is vertical.
        """
        self.position = position
        if type(rocktypes) != tuple:
            rocktypes = tuple(rocktypes)
        self.rocktypes = rocktypes
        self.terminal = terminal

    def __hash__(self):
        return hash((self.position, self.rocktypes, self.terminal))
    def __eq__(self, other):
        if isinstance(other, RSState):
            return self.position == other.position\
                and self.rocktypes == other.rocktypes\
                and self.terminal == other.terminal
        else:
            return False

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "RSState(%s | %s | %s)" % (str(self.position), str(self.rocktypes), str(self.terminal))

cdef class RSAction(Action):
    def __init__(self, name):
        self.name = name
    def __hash__(self):
        return hash(self.name)
    def __eq__(self, other):
        if isinstance(other, Action):
            return self.name == other.name
        elif type(other) == str:
            return self.name == other
    def __str__(self):
        return self.name
    def __repr__(self):
        return "Action(%s)" % self.name


cdef class MoveAction(RSAction):
    EAST = (1, 0)  # x is horizontal; x+ is right. y is vertical; y+ is up.
    WEST = (-1, 0)
    NORTH = (0, 1)
    SOUTH = (0, -1)
    cdef public tuple motion
    def __init__(self, motion):
        if motion not in {MoveAction.EAST, MoveAction.WEST,
                          MoveAction.NORTH, MoveAction.SOUTH}:
            raise ValueError("Invalid move motion %s" % motion)
        self.motion = motion
        super().__init__("move-%s" % str(motion))

MoveEast = MoveAction(MoveAction.EAST)
MoveWest = MoveAction(MoveAction.WEST)
MoveNorth = MoveAction(MoveAction.NORTH)
MoveSouth = MoveAction(MoveAction.SOUTH)

cdef class SampleAction(RSAction):
    def __init__(self):
        super().__init__("sample")

cdef class CheckAction(RSAction):
    cdef public int rock_id
    def __init__(self, rock_id):
        self.rock_id = rock_id
        super().__init__("check-%d" % self.rock_id)

cdef class RSObservation(Observation):
    cdef public str quality
    def __init__(self, quality):
        self.quality = quality
    def __hash__(self):
        return hash(self.quality)
    def __eq__(self, other):
        if isinstance(other, RSObservation):
            return self.quality == other.quality
        elif type(other) == str:
            return self.quality == other
    def __str__(self):
        return str(self.quality)
    def __repr__(self):
        return "RSObservation(%s)" % str(self.quality)

cdef class RSTransitionModel(TransitionModel):

    """ The model is deterministic """
    cdef public int _n
    cdef public dict _rock_locs
    cdef public _in_exit_area

    def __init__(self, n, rock_locs, in_exit_area):
        """
        rock_locs: a map from (x,y) location to rock_id
        in_exit_area: a function (x,y) -> Bool that returns True if (x,y) is in exit area"""
        self._n = n
        self._rock_locs = rock_locs
        self._in_exit_area = in_exit_area

    def _move_or_exit(self, position, action):
        expected = (position[0] + action.motion[0],
                    position[1] + action.motion[1])
        if self._in_exit_area(expected):
            return expected, True
        else:
            return (max(0, min(position[0] + action.motion[0], self._n-1)),
                    max(0, min(position[1] + action.motion[1], self._n-1))), False

    def probability(self, next_state, state, action, normalized=False, **kwargs):
        if next_state != self.sample(state, action):
            return EPSILON
        else:
            return 1.0 - EPSILON

    # cpdef sample(self, State state, Action action):
    #     next_state = copy.deepcopy(state)
    #     if state.terminal:
    #         return next_state  # already terminated. So no state transition happens
    #     if isinstance(action, MoveAction):
    #         next_state.position, exiting = self._move_or_exit(state.position, action)
    #         if exiting:
    #             next_state.terminal = True
    #     elif isinstance(action, SampleAction):
    #         position = next_state.position
    #         if position in self._rock_locs:
    #             rock_id = self._rock_locs[position]
    #             _rocktypes = list(next_state.rocktypes)
    #             _rocktypes[rock_id] = RockType.BAD
    #             next_state.rocktypes = tuple(_rocktypes)
    #     return next_state

    cpdef sample(self, State state, Action action):
        cdef State next_state
        cdef tuple next_position
        cdef tuple rocktypes, next_rocktypes
        cdef bint exiting = False
        cdef bint next_terminal = False
        cdef int rock_id
        cdef list _rocktypes

        next_position = tuple(state.position)
        rocktypes = tuple(state.rocktypes)
        next_rocktypes = rocktypes
        if state.terminal:
            next_terminal = True  # already terminated. So no state transition happens
        else:
            if isinstance(action, MoveAction):
                next_position, exiting = self._move_or_exit(state.position, action)
                if exiting:
                    next_terminal = True
            elif isinstance(action, SampleAction):
                if next_position in self._rock_locs:
                    rock_id = self._rock_locs[next_position]
                    _rocktypes = list(rocktypes)
                    _rocktypes[rock_id] = RockType.BAD
                    next_rocktypes = tuple(_rocktypes)
        return RSState(next_position, next_rocktypes, terminal=next_terminal)


    def argmax(self, state, action):
        """Returns the most likely next state"""
        return self.sample(state, action)


cdef class RSObservationModel(ObservationModel):

    cdef public float _half_efficiency_dist
    cdef public dict _rocks

    def __init__(self, rock_locs, half_efficiency_dist=20):
        self._half_efficiency_dist = half_efficiency_dist
        self._rocks = {rock_locs[pos]:pos for pos in rock_locs}

    def probability(self, observation, next_state, action):
        if isinstance(action, CheckAction):
            # compute efficiency
            rock_pos = self._rocks[action.rock_id]
            dist = euclidean_dist(rock_pos, next_state.position)
            eta = (1 + pow(2, -dist / self._half_efficiency_dist)) * 0.5

            # compute probability
            actual_rocktype = next_state.rocktypes[action.rock_id]
            if actual_rocktype == observation:
                return eta
            else:
                return 1.0 - eta
        else:
            if observation.quality is None:
                return 1.0 - EPSILON  # expected to receive no observation
            else:
                return EPSILON

    def sample(self, next_state, action, argmax=False):
        if not next_state.terminal and isinstance(action, CheckAction):
            # compute efficiency
            rock_pos = self._rocks[action.rock_id]
            dist = euclidean_dist(rock_pos, next_state.position)
            eta = (1 + pow(2, -dist / self._half_efficiency_dist)) * 0.5

            if argmax:
                keep = eta > 0.5
            else:
                keep = eta > random.uniform(0, 1)

            actual_rocktype = next_state.rocktypes[action.rock_id]
            if not keep:
                observed_rocktype = RockType.invert(actual_rocktype)
                return RSObservation(observed_rocktype)
            else:
                return RSObservation(actual_rocktype)
        else:
            # Terminated or not a check action. So no observation.
            return RSObservation(None)

    def argmax(self, next_state, action):
        """Returns the most likely observation"""
        return self.sample(next_state, action, argmax=True)


cdef class RSRewardModel(RewardModel):

    cdef public dict _rock_locs
    cdef public _in_exit_area

    def __init__(self, rock_locs, in_exit_area):
        self._rock_locs = rock_locs
        self._in_exit_area = in_exit_area

    def sample(self, state, action, next_state, normalized=False, **kwargs):
        # deterministic
        if state.terminal:
            return 0  # terminated. No reward
        if isinstance(action, SampleAction):
            # need to check the rocktype in `state` because it has turned bad in `next_state`
            if state.position in self._rock_locs:
                if state.rocktypes[self._rock_locs[state.position]] == RockType.GOOD:
                    return 10
                else:
                    # No rock or bad rock
                    return -10
            else:
                return 0  # problem didn't specify penalty for sampling empty space.

        elif isinstance(action, MoveAction):
            if self._in_exit_area(next_state.position):
                return 10
        return 0

    def argmax(self, state, action, next_state, normalized=False, **kwargs):
        raise NotImplementedError

    def probability(self, reward, state, action, next_state, normalized=False, **kwargs):
        raise NotImplementedError


cdef class RSPolicyModel(RolloutPolicy):
    """Simple policy model according to problem description."""
    cdef public set _all_actions
    def __init__(self, k):
        check_actions = set({CheckAction(rock_id) for rock_id in range(k)})
        self._all_actions = {MoveEast, MoveWest, MoveNorth, MoveSouth}
                             # SampleAction()} | check_actions

    def sample(self, state, normalized=False, **kwargs):
        return random.sample(self._all_actions, 1)[0]

    def probability(self, action, state, normalized=False, **kwargs):
        raise NotImplementedError

    def argmax(self, state, normalized=False, **kwargs):
        """Returns the most likely reward"""
        raise NotImplementedError

    def get_all_actions(self, **kwargs):
        return self._all_actions

    def rollout(self, state, history=None):
        return random.sample(self._all_actions, 1)[0]


cdef class RockSampleProblem(POMDP):

    cdef public int _n
    cdef public int _k
    cdef public dict _rock_locs

    def __init__(self, n, k, init_state, rock_locs, init_belief):
        self._n, self._k = n, k
        agent = Agent(init_belief,
                      RSPolicyModel(k),
                      RSTransitionModel(n, rock_locs, self.in_exit_area),
                      RSObservationModel(rock_locs),
                      RSRewardModel(rock_locs, self.in_exit_area))
        env = Environment(init_state,
                          RSTransitionModel(n, rock_locs, self.in_exit_area),
                          RSRewardModel(rock_locs, self.in_exit_area))
        self._rock_locs = rock_locs
        super().__init__(agent, env, name="RockSampleProblem")

    @staticmethod
    def random_free_location(n, not_free_locs):
        """returns a random (x,y) location in nxn grid that is free."""
        while True:
            loc = (random.randint(0, n-1),
                   random.randint(0, n-1))
            if loc not in not_free_locs:
                return loc

    def in_exit_area(self, pos):
        return pos[0] == self._n

    @staticmethod
    def generate_instance(n, k):
        """Returns init_state and rock locations for an instance of RockSample(n,k)"""

        rover_position = (0, random.randint(0, n-1))
        rock_locs = {}  # map from rock location to rock id
        for i in range(k):
            loc = RockSampleProblem.random_free_location(n, set(rock_locs.keys()) | set({rover_position}))
            rock_locs[loc] = i

        # random rocktypes
        rocktypes = tuple(RockType.random() for i in range(k))

        # Ground truth state
        init_state = RSState(rover_position, rocktypes, False)

        return init_state, rock_locs

    def print_state(self):
        string = "\n______ID______\n"
        rover_position = self.env.state.position
        rocktypes = self.env.state.rocktypes
        # Rock id map
        for y in range(self._n):
            for x in range(self._n+1):
                char = "."
                if x == self._n:
                    char = ">"
                if (x,y) in self._rock_locs:
                    char = str(self._rock_locs[(x,y)])
                if (x,y) == rover_position:
                    char = "R"
                string += char
            string += "\n"
        string += "_____G/B_____\n"
        # Good/bad map
        for y in range(self._n):
            for x in range(self._n+1):
                char = "."
                if x == self._n:
                    char = ">"
                if (x,y) in self._rock_locs:
                    if rocktypes[self._rock_locs[(x,y)]] == RockType.GOOD:
                        char = "$"
                    else:
                        char = "x"
                if (x,y) == rover_position:
                    char = "R"
                string += char
            string += "\n"
        print(string)



def test_planner(rocksample, planner, nsteps=3, discount=0.95):
    gamma = 1.0
    total_reward = 0
    total_discounted_reward = 0
    for i in range(nsteps):
        print("==== Step %d ====" % (i+1))
        action = planner.plan(rocksample.agent)

        # pomdp_py.visual.visualize_pouct_search_tree(rocksample.agent.tree,
        #                                             max_depth=5, anonymize=False)

        true_state = copy.deepcopy(rocksample.env.state)
        print("Terminal? %s" % true_state.terminal)
        env_reward = rocksample.env.state_transition(action, execute=True)
        true_next_state = copy.deepcopy(rocksample.env.state)

        real_observation = rocksample.env.provide_observation(rocksample.agent.observation_model,
                                                              action)
        rocksample.agent.update_history(action, real_observation)
        planner.update(rocksample.agent, action, real_observation)
        total_reward += env_reward
        total_discounted_reward += env_reward * gamma
        gamma *= discount
        print("True state: %s" % true_state)
        print("Action: %s" % str(action))
        print("Observation: %s" % str(real_observation))
        print("Reward: %s" % str(env_reward))
        print("Reward (Cumulative): %s" % str(total_reward))
        print("Reward (Cumulative Discounted): %s" % str(total_discounted_reward))
        if isinstance(planner, pomdp_py.POUCT):
            print("__num_sims__: %d" % planner.last_num_sims)
            print("__plan_time__: %.5f" % planner.last_planning_time)
        if isinstance(planner, pomdp_py.PORollout):
            print("__best_reward__: %d" % planner.last_best_reward)
        print("World:")
        rocksample.print_state()

        if rocksample.in_exit_area(rocksample.env.state.position):
            break
    return total_reward, total_discounted_reward


def init_particles_belief(k, num_particles, init_state, belief="uniform"):
    num_particles = 200
    particles = []
    for _ in range(num_particles):
        if belief == "uniform":
            rocktypes = []
            for i in range(k):
                rocktypes.append(RockType.random())
            rocktypes = tuple(rocktypes)
        elif belief == "groundtruth":
            rocktypes = copy.deepcopy(init_state.rocktypes)
        particles.append(RSState(init_state.position, rocktypes, False))
    init_belief = pomdp_py.Particles(particles)
    return init_belief


def main():
    n, k = 3,1
    init_state, rock_locs = RockSampleProblem.generate_instance(n, k)
    init_state_copy = copy.deepcopy(init_state)

    belief = "uniform"

    # init belief (uniform), represented in particles;
    # We don't factor the state here; We are also not doing any action prior.
    init_belief = init_particles_belief(k, 200, init_state, belief=belief)

    rocksample = RockSampleProblem(n, k, init_state, rock_locs, init_belief)
    rocksample.print_state()

    print("*** Testing POMCP ***")
    pomcp = pomdp_py.POMCP(max_depth=20, discount_factor=0.95,
                           planning_time=1., exploration_const=20,
                           rollout_policy=rocksample.agent.policy_model)
    tt, ttd = test_planner(rocksample, pomcp, nsteps=100, discount=0.95)

if __name__ == '__main__':
    main()



"""Plotting utilties"""
import matplotlib.pyplot as plt

def plot_points(xvals, yvals, color=None,
                size=1.5, label=None, connected=True, style="--", linewidth=1.5,
                xlabel='x', ylabel='f(x)', loc="lower right"):
    if not connected:
        plt.scatter(xvals, yvals, s=size, c=color, label=label)
    else:
        plt.plot(xvals, yvals, style, linewidth=linewidth, label=label)
    # plt.axhline(y=0, color='k')
    # plt.axvline(x=0, color='k')
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.legend(loc=loc)

def save_plot(path, bbox_inches='tight'):
    plt.savefig(path, bbox_inches=bbox_inches)
    plt.close()


# Plot polygons with colors
def plot_polygons(verts, colors, ax=None, edgecolor=None):
    """
    `verts` is a sequence of ( verts0, verts1, ...) where verts_i is a sequence of
    xy tuples of vertices, or an equivalent numpy array of shape (nv, 2).

    `c` is a sequence of (color0, color1, ...) where color_i is a color,
    represented by a hex string (7 characters #xxxxxx).

    Creates a PolygonCollection object in the axis `ax`."""
    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot(1,1,1)
    pc = PolyCollection(verts)
    pc.set_edgecolor(edgecolor)
    pc.set_facecolor(colors)
    ax.add_collection(pc)
    ax.set_xlabel('X axis')
    ax.set_ylabel('Y axis')


def plot_line(ax, p1, p2,
              linewidth=1, color='black', zorder=0, alpha=1.0, linestyle="-"):
    p1x, p1y = p1
    p2x, p2y = p2
    line = lines.Line2D([p1x, p2x], [p1y, p2y],
                        linewidth=linewidth, color=color, zorder=zorder,
                        alpha=alpha, linestyle=linestyle)
    ax.add_line(line)

def plot_circle(ax, center, radius, color="blue",
                fill=False, zorder=0, linewidth=0,
                edgecolor=None, label_text=None,
                alpha=1.0, text_color="white"):
    px, py = center
    circ = plt.Circle((px, py), radius, facecolor=color, fill=fill,
                      zorder=zorder, linewidth=linewidth, edgecolor=edgecolor, alpha=alpha)
    ax.add_artist(circ)
    if label_text:
        text = ax.text(px, py, label_text, color=text_color,
                        ha='center', va='center', size=7, weight='bold')
        text.set_path_effects([path_effects.Stroke(linewidth=1, foreground='black'),
                               path_effects.Normal()])



from pomdp_py.framework.basics cimport Action, Agent, Observation, State

cdef class Planner:

    cpdef public plan(self, Agent agent)
    cpdef public update(self, Agent agent, Action real_action, Observation real_observation)
    



"""
Some particular implementations of the interface for convenience
"""
import pomdp_py
import random

class SimpleState(pomdp_py.State):
    """A SimpleState is a state that stores
    one piece of hashable data and the equality
    of two states of this kind depends just on
    this data"""
    def __init__(self, data):
        self.data = data
    def __hash__(self):
        return hash(self.data)
    def __eq__(self, other):
        if isinstance(other, SimpleState):
            return self.data == other.data
        return False
    def __ne__(self, other):
        return not self.__eq__(other)
    def __str__(self):
        return str(self.data)
    def __repr__(self):
        return "SimpleState({})".format(self.data)

class SimpleAction(pomdp_py.Action):
    """A SimpleAction is an action defined by a string name"""
    def __init__(self, name):
        self.name = name
    def __hash__(self):
        return hash(self.name)
    def __eq__(self, other):
        if isinstance(other, SimpleAction):
            return self.name == other.name
        return False
    def __ne__(self, other):
        return not self.__eq__(other)
    def __str__(self):
        return self.name
    def __repr__(self):
        return "SimpleAction({})".format(self.name)

class SimpleObservation(pomdp_py.Observation):
    """A SimpleObservation is an observation
    with a piece of hashable data that defines
    the equality."""
    def __init__(self, data):
        self.data = data
    def __hash__(self):
        return hash(self.data)
    def __eq__(self, other):
        if isinstance(other, SimpleObservation):
            return self.data == other.data
        return False
    def __ne__(self, other):
        return not self.__eq__(other)
    def __str__(self):
        return str(self.data)
    def __repr__(self):
        return "SimpleObservation({})".format(self.data)

class DetTransitionModel(pomdp_py.TransitionModel):
    """A DetTransitionModel is a deterministic transition model.
    A probability of 1 - epsilon is given for correct transition,
    and epsilon is given for incorrect transition."""
    def __init__(self, epsilon=1e-12):
        self.epsilon = epsilon

    def probability(self, next_state, state, action):
        if self.sample(state, action) == next_state:
            return 1.0 - self.epsilon
        else:
            return self.epsilon

    def sample(self, state, action):
        raise NotImplementedError


class DetObservationModel(pomdp_py.ObservationModel):
    """A DetTransitionModel is a deterministic transition model.
    A probability of 1 - epsilon is given for correct transition,
    and epsilon is given for incorrect transition."""
    def __init__(self, epsilon=1e-12):
        self.epsilon = epsilon

    def probability(self, observation, next_state, action):
        if self.sample(next_state, action) == observation:
            return 1.0 - self.epsilon
        else:
            return self.epsilon

    def sample(self, next_state, action):
        raise NotImplementedError


class DetRewardModel(pomdp_py.RewardModel):
    """A DetRewardModel is a deterministic reward model (the most typical kind)."""
    def reward_func(self, state, action, next_state):
        raise NotImplementedError

    def sample(self, state, action, next_state):
        # deterministic
        return self.reward_func(state, action, next_state)

    def argmax(self, state, action, next_state):
        return self.sample(state, action, next_state)

class UniformPolicyModel(pomdp_py.RolloutPolicy):
    def __init__(self, actions):
        self.actions = actions

    def sample(self, state, **kwargs):
        return random.sample(self.actions, 1)[0]

    def get_all_actions(self, state=None, history=None):
        return self.actions

    def rollout(self, state, history=None):
        return random.sample(self.actions, 1)[0]


# Tabular models
class TabularTransitionModel(pomdp_py.TransitionModel):
    """This tabular transition model is built given a dictionary that maps a tuple
    (state, action, next_state) to a probability.  This model assumes that the
    given `weights` is complete, that is, it specifies the probability of all
    state-action-nextstate combinations
    """
    def __init__(self, weights):
        self.weights = weights
        self._states = set()
        for s, _, sp in weights:
            self._states.add(s)
            self._states.add(sp)

    def probability(self, next_state, state, action):
        if (state, action, next_state) in self.weights:
            return self.weights[(state, action, next_state)]
        raise ValueError("The transition probability for"\
                         f"{(state, action, next_state)} is not defined")

    def sample(self, state, action):
        next_states = list(self._states)
        probs = [self.probability(next_state, state, action)
                 for next_state in next_states]
        return random.choices(next_states, weights=probs, k=1)[0]

    def get_all_states(self):
        return self._states


class TabularObservationModel(pomdp_py.ObservationModel):
    """This tabular observation model is built given a dictionary that maps a tuple
    (next_state, action, observation) to a probability.  This model assumes that the
    given `weights` is complete.
    """
    def __init__(self, weights):
        self.weights = weights
        self._observations = set()
        for _, _, z in weights:
            self._observations.add(z)

    def probability(self, observation, next_state, action):
        """observation is emitted from state"""
        if (next_state, action, observation) in self.weights:
            return self.weights[(next_state, action, observation)]
        elif (next_state, observation) in self.weights:
            return self.weights[(next_state, observation)]
        raise ValueError("The observation probability for"
                         f"{(next_state, action, observation)} or {(next_state, observation)}"
                         "is not defined")

    def sample(self, next_state, action):
        observations = list(self._observations)
        probs = [self.probability(observation, next_state, action)
                 for observation in observations]
        return random.choices(observations, weights=probs, k=1)[0]

    def get_all_observations(self):
        return self._observations


class TabularRewardModel(pomdp_py.RewardModel):
    """This tabular reward model is built given a dictionary that maps a state or a
    tuple (state, action), or (state, action, next_state) to a probability.  This
    model assumes that the given `rewards` is complete.
    """
    def __init__(self, rewards):
        self.rewards = rewards

    def sample(self, state, action, *args):
        if state in self.rewards:
            return self.rewards[state]
        elif (state, action) in self.rewards:
            return self.rewards[(state, action)]
        else:
            if len(args) > 0:
                next_state = args[0]
                if (state, action, next_state) in self.rewards:
                    return self.rewards[(state, action, next_state)]

            raise ValueError("The reward is undefined for"
                             f"state={state}, action={action}"
                             f"next_state={args}")



"""The classic Tiger problem.

This is a POMDP problem; Namely, it specifies both
the POMDP (i.e. state, action, observation space)
and the T/O/R for the agent as well as the environment.

The description of the tiger problem is as follows: (Quote from `POMDP:
Introduction to Partially Observable Markov Decision Processes
<https://cran.r-project.org/web/packages/pomdp/vignettes/POMDP.pdf>`_ by
Kamalzadeh and Hahsler )

A tiger is put with equal probability behind one
of two doors, while treasure is put behind the other one.
You are standing in front of the two closed doors and
need to decide which one to open. If you open the door
with the tiger, you will get hurt (negative reward).
But if you open the door with treasure, you receive
a positive reward. Instead of opening a door right away,
you also have the option to wait and listen for tiger noises. But
listening is neither free nor entirely accurate. You might hear the
tiger behind the left door while it is actually behind the right
door and vice versa.

States: tiger-left, tiger-right
Actions: open-left, open-right, listen
Rewards:
    +10 for opening treasure door. -100 for opening tiger door.
    -1 for listening.
Observations: You can hear either "tiger-left", or "tiger-right".

Note that in this example, the TigerProblem is a POMDP that
also contains the agent and the environment as its fields. In
general this doesn't need to be the case. (Refer to more complicated
examples.)

"""
from pomdp_py.framework.basics cimport *
from pomdp_py.algorithms.po_uct cimport *
from pomdp_py.representations.belief.histogram import *

import pomdp_py
import random
import numpy as np
import sys

def build_states(strings):
    return {TigerState(s) for s in strings}
def build_actions(strings):
    return {TigerAction(s) for s in strings}
def build_observations(strings):
    return {TigerObservation(s) for s in strings}

cdef class TigerState(State):
    cdef public str name
    def __init__(self, name):
        if name != "tiger-left" and name != "tiger-right":
            raise ValueError("Invalid state: %s" % name)
        self.name = name
    def __hash__(self):
        return hash(self.name)
    def __eq__(self, other):
        if isinstance(other, TigerState):
            return self.name == other.name
        elif type(other) == str:
            return self.name == other
    def __str__(self):
        return self.name
    def __repr__(self):
        return "TigerState(%s)" % self.name

cdef class TigerAction(Action):
    def __init__(self, name):
        if name != "open-left" and name != "open-right"\
           and name != "listen":
            raise ValueError("Invalid action: %s" % name)
        self.name = name
    def __hash__(self):
        return hash(self.name)
    def __eq__(self, other):
        if isinstance(other, TigerAction):
            return self.name == other.name
        elif type(other) == str:
            return self.name == other
    def __str__(self):
        return self.name
    def __repr__(self):
        return "TigerAction(%s)" % self.name

cdef class TigerObservation(Observation):
    cdef public str name
    def __init__(self, name):
        if name != "tiger-left" and name != "tiger-right":
            raise ValueError("Invalid action: %s" % name)
        self.name = name
    def __hash__(self):
        return hash(self.name)
    def __eq__(self, other):
        if isinstance(other, TigerObservation):
            return self.name == other.name
        elif type(other) == str:
            return self.name == other
    def __str__(self):
        return self.name
    def __repr__(self):
        return "TigerObservation(%s)" % self.name

# Observation model
cdef class TigerObservationModel(ObservationModel):
    """This problem is small enough for the probabilities to be directly given
    externally"""
    cdef dict _probs
    def __init__(self, probs):
        self._probs = probs

    def probability(self, observation, next_state, action, normalized=False, **kwargs):
        return self._probs[next_state][action][observation]

    cpdef sample(self, State next_state, Action action, normalized=False):
        return self.get_distribution(next_state, action).random()

    def argmax(self, next_state, action, normalized=False, **kwargs):
        """Returns the most likely observation"""
        return max(self._probs[next_state][action], key=self._probs[next_state][action].get)

    cpdef get_distribution(self, State next_state, Action action):
        """Returns the underlying distribution of the model; In this case, it's just a histogram"""
        return Histogram(self._probs[next_state][action])

    def get_all_observations(self):
        return TigerProblem.OBSERVATIONS

# Transition Model
cdef class TigerTransitionModel(TransitionModel):
    """This problem is small enough for the probabilities to be directly given
            externally"""
    cdef dict _probs
    def __init__(self, probs):
        self._probs = probs

    def probability(self, next_state, state, action, normalized=False, **kwargs):
        return self._probs[state][action][next_state]

    cpdef sample(self, State state, Action action, normalized=False):
        return self.get_distribution(state, action).random()

    def argmax(self, state, action, normalized=False, **kwargs):
        """Returns the most likely next state"""
        return max(self._probs[state][action], key=self._probs[state][action].get)

    cpdef get_distribution(self, State state, Action action):
        """Returns the underlying distribution of the model"""
        return Histogram(self._probs[state][action])

    def get_all_states(self):
        return TigerProblem.STATES

# Reward Model
cdef class TigerRewardModel(RewardModel):
    cdef int _scale
    def __init__(self, scale=1):
        self._scale = scale
    cpdef _reward_func(self, State state, Action action):
        cdef int reward = 0
        if action == "open-left":
            if state== "tiger-right":
                reward += 10 * self._scale
            else:
                reward -= 100 * self._scale
        elif action == "open-right":
            if state== "tiger-left":
                reward += 10 * self._scale
            else:
                reward -= 100 * self._scale
        elif action == "listen":
            reward -= 1 * self._scale
        return reward

    def probability(self, reward, state, action, next_state, normalized=False):
        if reward == self._reward_func(state, action):
            return 1.0
        else:
            return 0.0

    cpdef sample(self, State state, Action action, State next_state, normalized=False):
        # deterministic
        return self._reward_func(state, action)

    def argmax(self, state, action, next_state, normalized=False):
        """Returns the most likely reward"""
        return self._reward_func(state, action)

    def get_distribution(self, state, action, next_state):
        """Returns the underlying distribution of the model"""
        reward = self._reward_func(state, action)
        return Histogram({reward:1.0})

# Policy Model
cdef class TigerPolicyModel(RandomRollout):
    """This is an extremely dumb policy model; To keep consistent
    with the framework."""
    def probability(self, action, state, normalized=False, **kwargs):
        raise NotImplementedError  # Never used

    def sample(self, state, normalized=False, **kwargs):
        return self.get_all_actions().random()

    def argmax(self, state, normalized=False, **kwargs):
        """Returns the most likely reward"""
        raise NotImplementedError

    def get_all_actions(self, **kwargs):
        return TigerProblem.ACTIONS


class TigerProblem(POMDP):

    STATES = build_states({"tiger-left", "tiger-right"})
    ACTIONS = build_actions({"open-left", "open-right", "listen"})
    OBSERVATIONS = build_observations({"tiger-left", "tiger-right"})

    def __init__(self, obs_probs, trans_probs, init_true_state, init_belief):
        """init_belief is a Distribution."""
        self._obs_probs = obs_probs
        self._trans_probs = trans_probs


        agent = Agent(init_belief,
                               TigerPolicyModel(),
                               TigerTransitionModel(self._trans_probs),
                               TigerObservationModel(self._obs_probs),
                               TigerRewardModel())
        env = Environment(init_true_state,
                                   TigerTransitionModel(self._trans_probs),
                                   TigerRewardModel())
        super().__init__(agent, env, name="TigerProblem")


def test_planner(tiger_problem, planner, nsteps=3):
    """
    Runs the action-feedback loop of Tiger problem POMDP

    Args:
        tiger_problem (TigerProblem): an instance of the tiger problem.
        planner (Planner): a planner
        nsteps (int): Maximum number of steps to run this loop.
    """
    for i in range(nsteps):
        action = planner.plan(tiger_problem.agent)
        print("==== Step %d ====" % (i+1))
        print("True state: %s" % tiger_problem.env.state)
        print("Belief: %s" % str(tiger_problem.agent.cur_belief))
        print("Action: %s" % str(action))
        print("Reward: %s" % str(tiger_problem.env.reward_model.sample(tiger_problem.env.state, action, None)))

        # Let's create some simulated real observation; Update the belief
        # Creating true observation for sanity checking solver behavior.
        # In general, this observation should be sampled from agent's observation model.
        real_observation = TigerObservation(tiger_problem.env.state.name)
        print(">> Observation: %s" % real_observation)
        tiger_problem.agent.update_history(action, real_observation)

        planner.update(tiger_problem.agent, action, real_observation)
        if isinstance(planner, POUCT):
            print("Num sims: %d" % planner.last_num_sims)
            print("Plan time: %.5f" % planner.last_planning_time)
        if isinstance(tiger_problem.agent.cur_belief, Histogram):
            new_belief = update_histogram_belief(tiger_problem.agent.cur_belief,
                                                          action, real_observation,
                                                          tiger_problem.agent.observation_model,
                                                          tiger_problem.agent.transition_model)
            tiger_problem.agent.set_belief(new_belief)

def build_setting(setting):
    result = {'obs_probs':{}, 'trans_probs':{}}
    for sp in setting['obs_probs']:
        result['obs_probs'][TigerState(sp)] = {}
        for a in setting['obs_probs'][sp]:
            result['obs_probs'][TigerState(sp)][TigerAction(a)] =\
                {TigerObservation(o):setting['obs_probs'][sp][a][o]
                 for o in setting['obs_probs'][sp][a]}
    for s in setting['trans_probs']:
        result['trans_probs'][TigerState(s)] = {}
        for a in setting['trans_probs'][s]:
            result['trans_probs'][TigerState(s)][TigerAction(a)] =\
                {TigerState(sp):setting['trans_probs'][s][a][sp]
                 for sp in setting['trans_probs'][s][a]}
    return result

def main():
    ## Setting 1:
    ## The values are set according to the paper.
    setting1 = {
        "obs_probs": {  # next_state -> action -> observation
            "tiger-left":{
                "open-left": {"tiger-left": 0.5, "tiger-right": 0.5},
                "open-right": {"tiger-left": 0.5, "tiger-right": 0.5},
                "listen": {"tiger-left": 0.85, "tiger-right": 0.15}
            },
            "tiger-right":{
                "open-left": {"tiger-left": 0.5, "tiger-right": 0.5},
                "open-right": {"tiger-left": 0.5, "tiger-right": 0.5},
                "listen": {"tiger-left": 0.15, "tiger-right": 0.85}
            }
        },

        "trans_probs": {  # state -> action -> next_state
            "tiger-left":{
                "open-left": {"tiger-left": 0.5, "tiger-right": 0.5},
                "open-right": {"tiger-left": 0.5, "tiger-right": 0.5},
                "listen": {"tiger-left": 1.0, "tiger-right": 0.0}
            },
            "tiger-right":{
                "open-left": {"tiger-left": 0.5, "tiger-right": 0.5},
                "open-right": {"tiger-left": 0.5, "tiger-right": 0.5},
                "listen": {"tiger-left": 0.0, "tiger-right": 1.0}
            }
        }
    }

    ## Setting 2:
    ## Based on my understanding of T and O; Treat the state as given.
    setting2 = {
        "obs_probs": {  # next_state -> action -> observation
            "tiger-left":{
                "open-left": {"tiger-left": 1.0, "tiger-right": 0.0},
                "open-right": {"tiger-left": 1.0, "tiger-right": 0.0},
                "listen": {"tiger-left": 0.85, "tiger-right": 0.15}
            },
            "tiger-right":{
                "open-left": {"tiger-left": 0.0, "tiger-right": 1.0},
                "open-right": {"tiger-left": 0.0, "tiger-right": 1.0},
                "listen": {"tiger-left": 0.15, "tiger-right": 0.85}
            }
        },

        "trans_probs": {  # state -> action -> next_state
            "tiger-left":{
                "open-left": {"tiger-left": 1.0, "tiger-right": 0.0},
                "open-right": {"tiger-left": 1.0, "tiger-right": 0.0},
                "listen": {"tiger-left": 1.0, "tiger-right": 0.0}
            },
            "tiger-right":{
                "open-left": {"tiger-left": 0.0, "tiger-right": 1.0},
                "open-right": {"tiger-left": 0.0, "tiger-right": 1.0},
                "listen": {"tiger-left": 0.0, "tiger-right": 1.0}
            }
        }
    }

    # setting1 resets the problem after the agent chooses open-left or open-right;
    # It's reasonable - when the agent opens one door and sees a tiger, it gets
    # killed; when the agent sees a treasure, it will take some of the treasure
    # and leave. Thus, the agent will have no incentive to close the door and do
    # this again. The setting2 is the case where the agent is a robot, and only
    # cares about getting higher reward.
    setting = build_setting(setting1)

    init_true_state = random.choice(list(TigerProblem.STATES))
    init_belief = Histogram({TigerState("tiger-left"): 0.5,
                                      TigerState("tiger-right"): 0.5})
    tiger_problem = TigerProblem(setting['obs_probs'],
                                 setting['trans_probs'],
                                 init_true_state, init_belief)
    print("\n** Testing POUCT **")
    pouct = POUCT(max_depth=10, discount_factor=0.95,
                           num_sims=4096, exploration_const=110,
                           rollout_policy=tiger_problem.agent.policy_model)
    test_planner(tiger_problem, pouct, nsteps=10)


if __name__ == '__main__':
    main()



import pomdp_py
import random
from pomdp_problems.tag.domain.action import *
from pomdp_problems.tag.models.transition_model import *

class TagPolicyModel(pomdp_py.RolloutPolicy):
    def __init__(self, grid_map=None):
        self._grid_map = grid_map

    def sample(self, state, **kwargs):
        return random.sample(self.get_all_actions(state, **kwargs), 1)[0]
    
    def get_all_actions(self, state=None, history=None):
        if state is not None:
            if self._grid_map is not None:
                valid_motions = self._grid_map.valid_motions(state.robot_position,
                                                             all_motions=MOTION_ACTIONS)
                return valid_motions | set({TagAction()})
        return MOTION_ACTIONS | set({TagAction()})

    def rollout(self, state, history=None):
        return random.sample(self.get_all_actions(state=state, history=history), 1)[0]




"""
Provides utility to convert a POMDP written in pomdp_py to
a POMDP file format (.pomdp or .pomdpx). Output to a file.
"""
import subprocess
import os
import pomdp_py
import numpy as np
import xml.etree.ElementTree as ET

def to_pomdp_file(agent, output_path=None,
                  discount_factor=0.95):
    """
    Pass in an Agent, and use its components to generate
    a .pomdp file to `output_path`.

    The .pomdp file format is specified at:
    http://www.pomdp.org/code/pomdp-file-spec.html

    Note:

    * It is assumed that the reward is independent of the observation.
    * The state, action, and observations of the agent must be
      explicitly enumerable.
    * The state, action and observations of the agent must be
      convertable to a string that does not contain any blank space.

    Args:
        agent (~pomdp_py.framework.basics.Agent): The agent
        output_path (str): The path of the output file to write in. Optional.
                           Default None.
        discount_factor (float): The discount factor
    Returns:
        (list, list, list): The list of states, actions, observations that
           are ordered in the same way as they are in the .pomdp file.
    """
    # Preamble
    try:
        all_states = list(agent.all_states)
        all_actions = list(agent.all_actions)
        all_observations = list(agent.all_observations)
    except NotImplementedError:
        raise ValueError("S, A, O must be enumerable for a given agent to convert to .pomdp format")

    content = "discount: %f\n" % discount_factor
    content += "values: reward\n" # We only consider reward, not cost.

    list_of_states = " ".join(str(s) for s in all_states)
    assert len(list_of_states.split(" ")) == len(all_states),\
        "states must be convertable to strings without blank spaces"
    content += "states: %s\n" % list_of_states

    list_of_actions = " ".join(str(a) for a in all_actions)
    assert len(list_of_actions.split(" ")) == len(all_actions),\
        "actions must be convertable to strings without blank spaces"
    content += "actions: %s\n" % list_of_actions

    list_of_observations = " ".join(str(o) for o in all_observations)
    assert len(list_of_observations.split(" ")) == len(all_observations),\
        "observations must be convertable to strings without blank spaces"
    content += "observations: %s\n" % list_of_observations

    # Starting belief state - they need to be normalized
    total_belief = sum(agent.belief[s] for s in all_states)
    content += "start: %s\n" % (" ".join(["%f" % (agent.belief[s]/total_belief)
                                          for s in all_states]))

    # State transition probabilities - they need to be normalized
    for s in all_states:
        for a in all_actions:
            probs = []
            for s_next in all_states:
                prob = agent.transition_model.probability(s_next, s, a)
                probs.append(prob)
            total_prob = sum(probs)
            for i, s_next in enumerate(all_states):
                prob_norm = probs[i] / total_prob
                content += 'T : %s : %s : %s %f\n' % (a, s, s_next, prob_norm)

    # Observation probabilities - they need to be normalized
    for s_next in all_states:
        for a in all_actions:
            probs = []
            for o in all_observations:
                prob = agent.observation_model.probability(o, s_next, a)
                probs.append(prob)
            total_prob = sum(probs)
            assert total_prob > 0.0,\
                "No observation is probable under state={} action={}"\
                .format(s_next, a)
            for i, o in enumerate(all_observations):
                prob_norm = probs[i] / total_prob
                content += 'O : %s : %s : %s %f\n' % (a, s_next, o, prob_norm)

    # Immediate rewards
    for s in all_states:
        for a in all_actions:
            for s_next in all_states:
                # We will take the argmax reward, which works for deterministic rewards.
                r = agent.reward_model.sample(s, a, s_next)
                content += 'R : %s : %s : %s : *  %f\n' % (a, s, s_next, r)

    if output_path is not None:
        with open(output_path, "w") as f:
            f.write(content)
    return all_states, all_actions, all_observations



def to_pomdpx_file(agent, pomdpconvert_path,
                   output_path=None,
                   discount_factor=0.95):
    """
    Converts an agent to a pomdpx file. This works by first converting the agent into a .pomdp file and then
    using the :code:`pomdpconvert` utility program to convert that file to a .pomdpx file. Check out
    :code:`pomdpconvert` at `github://AdaCompNUS/sarsop <https://github.com/AdaCompNUS/sarsop>`_

    Follow the instructions at https://github.com/AdaCompNUS/sarsop
    to download and build sarsop (I tested on Ubuntu 18.04, gcc version 7.5.0)

    See documentation for pomdpx at:
    https://bigbird.comp.nus.edu.sg/pmwiki/farm/appl/index.php?n=Main.PomdpXDocumentation

    First converts the agent into .pomdp, then convert it to pomdpx.

    Args:
        agent (~pomdp_py.framework.basics.Agent): The agent
        pomdpconvert_path (str): Path to the :code:`pomdpconvert` binary
        output_path (str): The path of the output file to write in. Optional.
                           Default None.
        discount_factor (float): The discount factor
    """
    pomdp_path = "./temp-pomdp.pomdp"
    to_pomdp_file(agent, pomdp_path,
                  discount_factor=discount_factor)
    proc = subprocess.Popen([pomdpconvert_path, pomdp_path])
    proc.wait()

    pomdpx_path = pomdp_path + "x"
    assert os.path.exists(pomdpx_path), "POMDPx conversion failed."

    with open(pomdpx_path, 'r') as f:
        content = f.read()

    if output_path is not None:
        os.rename(pomdpx_path, output_path)

    # Delete temporary files
    os.remove(pomdp_path)
    if os.path.exists(pomdpx_path):
        os.remove(pomdpx_path)


def parse_pomdp_solve_output(alpha_path, pg_path=None):
    """Parse the output of pomdp_solve, given
    by an .alpha file and a .pg file.

    Given a path to a .alpha file, read and interpret its contents.
    The file formats are specified at:
    https://www.pomdp.org/code/alpha-file-spec.html
    https://www.pomdp.org/code/pg-file-spec.html

    Note on policy graph (from the official website): To use this first requires
    knowing which of the policy graph states to start in. This can be achieved
    by finding the alpha vector with the maximal dot product with the initial
    starting state.

    Note: Parsing the .alpha file is required. The .pg path is optional (this
        is because I noticed some errors in the .pg file produced)

    Returns:
        alphas: [(alpha_vector, action_number) ...]
        policy_graph: a mapping from node number to (action_number, edges)
    """
    alphas = []  # (alpha_vector, action_number) tuples
    with open(alpha_path, 'r') as f:
        action_number = None
        alpha_vector = None
        mode = "action"
        for line in f:
            line = line.rstrip()
            if len(line) == 0:
                continue

            if mode == "action":
                action_number = int(line)
                mode = "alpha"
            elif mode == "alpha":
                alpha_vector = tuple(map(float, line.split(" ")))
                alphas.append((alpha_vector, action_number))
                mode = "action"
                action_number = None
                alpha_vector = None

    policy_graph = {}  # a mapping from node number to (action_number, edges)
    if pg_path is None:
        return alphas
    else:
        with open(pg_path, 'r') as f:
            for line in f:
                line = line.rstrip()
                if len(line) == 0:
                    continue
                parts = list(map(int, line.split()))  # Splits on whitespace
                assert parts[0] not in policy_graph,\
                    "The node id %d already exists. Something wrong" % parts[0]
                policy_graph[parts[0]] = (parts[1], parts[2:])
        return alphas, policy_graph



class AlphaVectorPolicy(pomdp_py.Planner):
    """
    An offline POMDP policy is specified by a collection
    of alpha vectors, each associated with an action. When
    planning is needed, the dot product of these alpha vectors
    and the agent's belief vector is computed and the alpha
    vector leading to the maximum is the 'dominating' alpha
    vector and we return its corresponding action.

    An offline policy can be optionally represented as
    a policy graph. In this case, the agent can plan without
    actively maintaining a belief because the policy graph
    is a finite state machine that transitions by observations.

    This can be constructed using .policy file created by sarsop.
    """
    def __init__(self, alphas, states):
        """
        Args:
            alphas (list): A list of (alpha_vector, action) tuples.
                An alpha_vector is a list of floats :code:`[V1, V2, ... VN]`.
            states (list): List of states, ordered as in .pomdp file
        """
        self.alphas = alphas
        self.states = states

    def plan(self, agent):
        """Returns an action that is mapped by the agent belief, under this policy"""
        b = [agent.belief[s] for s in self.states]
        _, action = max(self.alphas,
                 key=lambda va: np.dot(b, va[0]))
        return action

    def value(self, belief):
        """
        Returns the value V(b) under this alpha vector policy.

        :math:`V(b) = max_{a\in\Gamma} {a} \cdot b`
        """
        b = [belief[s] for s in self.states]
        alpha_vector, _ = max(self.alphas,
                              key=lambda va: np.dot(b, va[0]))
        return np.dot(b, alpha_vector)

    @classmethod
    def construct(cls, policy_path,
                  states, actions, solver="pomdpsol"):
        """
        Returns an AlphaVectorPolicy, given `alphas`,
        which are the output of parse_appl_policy_file.

        Args:
            policy_path (str): Path to the generated .policy file
                (for sarsop) or .alpha file (for pomdp-solve)
            states (list): A list of States, in the same order as in the .pomdp file
            actions (list): A list of Actions, in the same order as in the .pomdp file
        Returns:
            AlphaVectorPolicy: The policy stored in the given policy file.
        """
        if solver == "pomdp-solve" or solver == "vi":
            return cls.construct_from_pomdp_solve(policy_path, states, actions)
        elif solver == "pomdpsol" or solver == "sarsop":
            alphas = []
            root = ET.parse(policy_path).getroot()
            for vector in root.find("AlphaVector").iter("Vector"):
                action = actions[int(vector.attrib["action"])]
                alpha_vector = tuple(map(float, vector.text.split()))
                alphas.append((alpha_vector, action))
            return AlphaVectorPolicy(alphas,
                                     states)

    @classmethod
    def construct_from_pomdp_solve(cls, alpha_path,
                                   states, actions):
        alphas_with_action_numbers = parse_pomdp_solve_output(alpha_path)
        alphas = [(alpha_vector, actions[action_number])
                  for alpha_vector, action_number in alphas_with_action_numbers]
        return AlphaVectorPolicy(alphas, states)


class PGNode:
    """A node on the policy graph"""
    def __init__(self, node_id, alpha_vector, action):
        self.node_id = node_id
        self.alpha_vector = alpha_vector
        self.action = action
    def __eq__(self, other):
        if isinstance(other, PolicyNode):
            return self.node_id == other.node_id
        return False
    def __hash__(self):
        return hash(self.node_id)
    def __str__(self):
        return repr(self)
    def __repr__(self):
        return "NodeID(%d)::AlphaVector(%s)::Action(%s)\n"\
            % (self.node_id, str(self.alpha_vector), self.action)


class PolicyGraph(pomdp_py.Planner):
    """A PolicyGraph encodes a POMDP plan. It
    can be constructed from the alphas and policy graph
    format output by Cassandra's pomdp-solver."""

    def __init__(self, nodes, edges, states):
        """
        Initializes a PolicyGraph.
        Args:
            nodes (list): A list of PGNodes
            edges (dict): Mapping from node_id to a dictionary {observation -> node_id}
            states (list): List of states, ordered as in .pomdp file
        """
        self.nodes = {n.node_id:n for n in nodes}
        self.edges = edges
        self.states = states
        self._current_node = None

    @classmethod
    def construct(cls, alpha_path, pg_path,
                  states, actions, observations):
        """
        See parse_pomdp_solve_output for detailed definitions of
        alphas and pg.

        Args:
            alpha_path (str): Path to .alpha file
            pg_path (str): Path to .pg file
            states (list): List of states, ordered as in .pomdp file
            actions (list): List of actions, ordered as in .pomdp file
            observations (list): List of observations, ordered as in .pomdp file
        """
        # alphas (list): List of ( [V1, V2, ... VN], A ) tuples
        # pg (dict): {node_id -> (A, edges)}
        alphas, pg = parse_pomdp_solve_output(alpha_path, pg_path)

        nodes = []
        for node_id, (alpha_vector, action_number) in enumerate(alphas):
            node = PGNode(node_id, alpha_vector, actions[action_number])
            nodes.append(node)

        edges = {}
        for node_id in pg:
            assert 0 <= node_id < len(nodes), "Invalid node id in policy graph"

            action_number, o_links = pg[node_id]
            assert actions[action_number] == nodes[node_id].action,\
                "Inconsistent action mapping"

            edges[node_id] = {}

            for o_index, next_node_id in enumerate(o_links):
                observation = observations[o_index]
                edges[node_id][observation] = next_node_id
        return PolicyGraph(nodes, edges, states)

    def plan(self, agent):
        """Returns an action that is mapped by the agent belief, under this policy"""
        if self._current_node is None:
            self._current_node = self._find_node(agent)
        return self._current_node.action

    def _find_node(self, agent):
        """Locate the node in the policy graph corresponding to the agent's current
        belief state. """
        b = [agent.belief[s] for s in self.states]
        nid = max(self.nodes,
                   key=lambda nid: np.dot(b, self.nodes[nid].alpha_vector))
        return self.nodes[nid]

    def update(self, agent, action, observation):
        """
        Updates the planner based on real action and observation.
        Basically sets the current node pointer based on the incoming
        observation."""
        if self._current_node is None:
            # Find out the node number using agent current belief
            self._current_node = self._find_node(agent)
        # Transition the current node following the graph
        self._current_node = self.nodes[self.edges[self._current_node.node_id][observation]]



"""
Defines the Action for the 2D Multi-Object Search domain;

Action space: 

    Motion :math:`\cup` Look :math:`\cup` Find

* Motion Actions scheme 1: South, East, West, North.
* Motion Actions scheme 2: Left 45deg, Right 45deg, Forward
* Look: Interprets sensor input as observation
* Find: Marks objects observed in the last Look action as
  (differs from original paper; reduces action space)

It is possible to force "Look" after every N/S/E/W action;
then the Look action could be dropped. This is optional behavior.
"""
import pomdp_py
import math

###### Actions ######
class Action(pomdp_py.Action):
    """Mos action; Simple named action."""
    def __init__(self, name):
        self.name = name
    def __hash__(self):
        return hash(self.name)
    def __eq__(self, other):
        if isinstance(other, Action):
            return self.name == other.name
        elif type(other) == str:
            return self.name == other
    def __str__(self):
        return self.name
    def __repr__(self):
        return "Action(%s)" % self.name

MOTION_SCHEME="xy"  # can be either xy or vw
STEP_SIZE=1
class MotionAction(Action):
    # scheme 1 (vx,vy,th)
    SCHEME_XYTH = "xyth"
    EAST = (STEP_SIZE, 0, 0)  # x is horizontal; x+ is right. y is vertical; y+ is down.
    WEST = (-STEP_SIZE, 0, math.pi)
    NORTH = (0, -STEP_SIZE, 3*math.pi/2)
    SOUTH = (0, STEP_SIZE, math.pi/2)
    
    # scheme 2 (vt, vw) translational, rotational velocities.
    SCHEME_VW = "vw" 
    FORWARD = (STEP_SIZE, 0)
    BACKWARD = (-STEP_SIZE, 0)
    LEFT = (0, -math.pi/4)  # left 45 deg
    RIGHT = (0, math.pi/4) # right 45 deg

    # scheme 3 (vx,vy)
    SCHEME_XY = "xy"
    EAST2D = (STEP_SIZE, 0)  # x is horizontal; x+ is right. y is vertical; y+ is down.
    WEST2D = (-STEP_SIZE, 0)
    NORTH2D = (0, -STEP_SIZE)
    SOUTH2D = (0, STEP_SIZE)

    SCHEMES = {"xyth", "xy", "vw"}

    def __init__(self, motion,
                 scheme=MOTION_SCHEME, distance_cost=1,
                 motion_name=None):
        """
        motion (tuple): a tuple of floats that describes the motion;
        scheme (str): description of the motion scheme; Either
                      "xy" or "vw"
        """
        if scheme not in MotionAction.SCHEMES:
            raise ValueError("Invalid motion scheme %s" % scheme)

        if scheme == MotionAction.SCHEME_XYTH:
            if motion not in {MotionAction.EAST, MotionAction.WEST,
                              MotionAction.NORTH, MotionAction.SOUTH}:
                raise ValueError("Invalid move motion %s" % str(motion))
        elif scheme == MotionAction.SCHEME_VW:
            if motion not in {MotionAction.FORWARD, MotionAction.BACKWARD,
                              MotionAction.LEFT, MotionAction.RIGHT}:
                raise ValueError("Invalid move motion %s" % str(motion))
        elif scheme == MotionAction.SCHEME_XY:
            if motion not in {MotionAction.EAST2D, MotionAction.WEST2D,
                              MotionAction.NORTH2D, MotionAction.SOUTH2D}:
                raise ValueError("Invalid move motion %s" % str(motion))
            
        self.motion = motion
        self.scheme = scheme
        self.distance_cost = distance_cost
        if motion_name is None:
            motion_name = str(motion)
        super().__init__("move-%s-%s" % (scheme, motion_name))
        
# Define some constant actions
MoveEast = MotionAction(MotionAction.EAST,   scheme=MotionAction.SCHEME_XYTH, motion_name="East")
MoveWest = MotionAction(MotionAction.WEST,   scheme=MotionAction.SCHEME_XYTH, motion_name="West")
MoveNorth = MotionAction(MotionAction.NORTH, scheme=MotionAction.SCHEME_XYTH, motion_name="North")
MoveSouth = MotionAction(MotionAction.SOUTH, scheme=MotionAction.SCHEME_XYTH, motion_name="South")

MoveForward  = MotionAction(MotionAction.FORWARD,  scheme=MotionAction.SCHEME_VW, motion_name="Forward")
MoveBackward = MotionAction(MotionAction.BACKWARD, scheme=MotionAction.SCHEME_VW, motion_name="Backward")
MoveLeft     = MotionAction(MotionAction.LEFT,     scheme=MotionAction.SCHEME_VW, motion_name="TurnLeft")
MoveRight    = MotionAction(MotionAction.RIGHT,    scheme=MotionAction.SCHEME_VW, motion_name="TurnRight")

MoveEast2D  = MotionAction(MotionAction.EAST2D,  scheme=MotionAction.SCHEME_XY, motion_name="East2D")
MoveWest2D  = MotionAction(MotionAction.WEST2D,  scheme=MotionAction.SCHEME_XY, motion_name="West2D")
MoveNorth2D = MotionAction(MotionAction.NORTH2D, scheme=MotionAction.SCHEME_XY, motion_name="North2D")
MoveSouth2D = MotionAction(MotionAction.SOUTH2D, scheme=MotionAction.SCHEME_XY, motion_name="South2D")


class LookAction(Action):
    # For simplicity, this LookAction is not parameterized by direction
    def __init__(self):
        super().__init__("look")
        
class FindAction(Action):
    def __init__(self):
        super().__init__("find")

Look = LookAction()
Find = FindAction()

if MOTION_SCHEME == "xy":
    ALL_MOTION_ACTIONS = {MoveEast, MoveWest, MoveNorth, MoveSouth}
elif MOTION_SCHEME == "vw":
    ALL_MOTION_ACTIONS = {MoveForward, MoveBackward, MoveLeft, MoveRight}
else:
    raise ValueError("motion scheme '%s' is invalid" % MOTION_SCHEME)



"""
Utilties for typography, i.e. dealing with
strings for the purpose of displaying them.
"""

# Colors on terminal https://stackoverflow.com/a/287944/2893053
class bcolors:
    WHITE = '\033[97m'
    CYAN = '\033[96m'
    MAGENTA = '\033[95m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    ENDC = '\033[0m'

    @staticmethod
    def disable():
        bcolors.WHITE   = ''
        bcolors.CYAN    = ''
        bcolors.MAGENTA = ''
        bcolors.BLUE    = ''
        bcolors.GREEN   = ''
        bcolors.YELLOW  = ''
        bcolors.RED     = ''
        bcolors.ENDC    = ''

    @staticmethod
    def s(color, content):
        """Returns a string with color when shown on terminal.
        `color` is a constant in `bcolors` class."""
        return color + content + bcolors.ENDC

def info(content):
    return bcolors.s(bcolors.BLUE, content)

def note(content):
    return bcolors.s(bcolors.YELLOW, content)

def error(content):
    return bcolors.s(bcolors.GREEN, content)

def warning(content):
    return bcolors.s(bcolors.YELLOW, content)

def success(content):
    return bcolors.s(bcolors.GREEN, content)

def bold(content):
    return bcolors.s(bcolors.BOLD, content)

def white(content):
    return bcolors.s(bcolors.WHITE, content)

def green(content):
    return bcolors.s(bcolors.GREEN, content)

def cyan(content):
    return bcolors.s(bcolors.CYAN, content)

def magenta(content):
    return bcolors.s(bcolors.MAGENTA, content)

def blue(content):
    return bcolors.s(bcolors.BLUE, content)

def green(content):
    return bcolors.s(bcolors.GREEN, content)

def yellow(content):
    return bcolors.s(bcolors.YELLOW, content)

def red(content):
    return bcolors.s(bcolors.RED, content)

def white(content):
    return bcolors.s(bcolors.WHITE, content)

colors = {
    "white",
    "green",
    "cyan",
    "magenta",
    "blue",
    "green",
    "yellow",
    "red",
    "white",
}



"""Plot the light dark environment"""
import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection
import pomdp_problems.light_dark as ld
from pomdp_py.utils import plotting, colors
from pomdp_py.utils.misc import remap

class LightDarkViz:
    """This class deals with visualizing a light dark domain"""

    def __init__(self, env, x_range, y_range, res):
        """
        Args:
            env (LightDarkEnvironment): Environment for light dark domain.
            x_range (tuple): a tuple of floats (x_min, x_max).
            y_range (tuple): a tuple of floats (y_min, y_max).
            res (float): specifies the size of each rectangular strip to draw;
                As in the paper, the light is at a location on the x axis.
        """
        self._env = env
        self._res = res
        self._x_range = x_range
        self._y_range = y_range
        fig = plt.gcf()
        self._ax = fig.add_subplot(1,1,1)
        self._goal_pos = None
        self._m_0 = None  # initial belief pose

        # For tracking the path; list of robot position tuples
        self._log_paths = {}

    def log_position(self, position, path=0):
        if path not in self._log_paths:
            self._log_paths[path] = []
        self._log_paths[path].append(position)

    def set_goal(self, goal_pos):
        self._goal_pos = goal_pos

    def set_initial_belief_pos(self, m_0):
        self._m_0 = m_0

    def plot(self,
             path_colors={0: [(0,0,0), (0,0,254)]},
             path_styles={0: "--"},
             path_widths={0: 1}):
        self._plot_gradient()
        self._plot_path(path_colors, path_styles, path_widths)
        self._plot_robot()
        self._plot_goal()
        self._plot_initial_belief_pos()

    def _plot_robot(self):
        cur_pos = self._env.state.position
        plotting.plot_circle(self._ax, cur_pos,
                         0.25, # tentative
                         color="black", fill=False,
                         linewidth=1, edgecolor="black",
                         zorder=3)

    def _plot_initial_belief_pos(self):
        if self._m_0 is not None:
            plotting.plot_circle(self._ax, self._m_0,
                                 0.25, # tentative
                                 color="black", fill=False,
                                 linewidth=1, edgecolor="black",
                                 zorder=3)

    def _plot_goal(self):
        if self._goal_pos is not None:
            plotting.plot_circle(self._ax,
                                 self._goal_pos,
                                 0.25,  # tentative
                                 linewidth=1, edgecolor="blue",
                                 zorder=3)

    def _plot_path(self, colors, styles, linewidths):
        """Plot robot path"""
        # Plot line segments
        for path in self._log_paths:
            if path not in colors:
                path_color = [(0,0,0)] * len(self._log_paths[path])
            else:
                if len(colors[path]) == 2:
                    c1, c2 = colors[path]
                    path_color = colors.linear_color_gradient(c1, c2,
                                                              len(self._log_paths[path]),
                                                              normalize=True)
                else:
                    path_color = [colors[path]] * len(self._log_paths[path])

            if path not in styles:
                path_style = "--"
            else:
                path_style = styles[path]

            if path not in linewidths:
                path_width = 1
            else:
                path_width = linewidths[path]

            for i in range(1, len(self._log_paths[path])):
                p1 = self._log_paths[path][i-1]
                p2 = self._log_paths[path][i]
                try:
                    plotting.plot_line(self._ax, p1, p2, color=path_color[i],
                                   linestyle=path_style, zorder=2, linewidth=path_width)
                except Exception:
                    import pdb; pdb.set_trace()

    def _plot_gradient(self):
        """display the light dark domain."""
        xmin, xmax = self._x_range
        ymin, ymax = self._y_range
        # Note that higher brightness has lower brightness value
        hi_brightness = self._env.const
        lo_brightness = max(0.5 * (self._env.light - xmin)**2 + self._env.const,
                            0.5 * (self._env.light - xmax)**2 + self._env.const)
        # Plot a bunch of rectangular strips along the x axis
        # Check out: https://stackoverflow.com/questions/10550477
        x = xmin
        verts = []
        colors = []
        while x < xmax:
            x_next = x + self._res
            verts.append([(x, ymin), (x_next, ymin), (x_next, ymax), (x, ymax)])
            # compute brightness based on equation in the paper
            brightness = 0.5 * (self._env.light - x)**2 + self._env.const
            # map brightness to a grayscale color
            grayscale = int(round(remap(brightness, hi_brightness, lo_brightness, 255, 0)))
            grayscale_hex = colors.rgb_to_hex((grayscale, grayscale, grayscale))
            colors.append(grayscale_hex)
            x = x_next
        plotting.plot_polygons(verts, colors, ax=self._ax)
        self._ax.set_xlim(xmin, xmax)
        self._ax.set_ylim(ymin, ymax)

if __name__ == "__main__":
    env = ld.LightDarkEnvironment(ld.State((0.5, 2.5)),  # init state
                                  (1.5, -1),  # goal pose
                                  5,  # light
                                  1)  # const
    viz = LightDarkViz(env, (-1, 7), (-2, 4), 0.1)
    viz.log_position((5,2))
    viz.log_position((5,0))
    viz.log_position((4,-1))
    viz.plot()

    plt.show()



"""Defines the TransitionModel for the continuous light-dark domain;

Origin: Belief space planning assuming maximum likelihood observations

Quote from the paper:

    The underlying system dynamics are linear with zero process noise,
    :math:`f(x_t,u_t)=x_t+u`. This means the transition dynamics is
    deterministic.
"""
import pomdp_py
import copy
import numpy as np


class TransitionModel(pomdp_py.TransitionModel):
    """
    The underlying deterministic system dynamics
    """
    def __init__(self, epsilon=1e-9):
        self._epsilon = epsilon

    def probability(self, next_state, state, action, **kwargs):
        """
        Deterministic.
        """
        expected_position = tuple(self.func(state.position, action))
        if next_state.position == expected_position:
            return 1.0 - self.epsilon
        else:
            return self.epsilon

    def sample(self, state, action):
        next_state = copy.deepcopy(state)
        next_state.position = tuple(self.func(state.position, action))
        return next_state

    def argmax(self, state, action):
        """Returns the most likely next state"""
        return self.sample(state, action)

    def func(self):
        """Returns the function of the underlying system dynamics.
        The function is: (xt, ut) -> xt+1 where xt, ut, xt+1 are
        all numpy arrays."""
        def f(xt, ut):
            return np.array([xt[0] + ut[0],
                             xt[1] + ut[1]])
        return f

    def jac_dx(self):
        """Returns the function of the jacobian of the system dynamics
        function with respect to the state vector mt: (mt, ut) -> At"""
        def dfdx(mt, ut):
            # The result of computing the jacobian by hand
            return np.array([[ut[0], mt[1] + ut[1]],
                             [mt[0] + ut[0], ut[1]]])
        return dfdx

    def jac_du(self):
        """Returns the function of the jacobian of the system dynamics
        function with respect to the state vector mt: (mt, ut) -> Bt"""
        def dfdu(mt, ut):
            # The result of computing the jacobian by hand
            return np.array([[mt[0], mt[1] + ut[1]],
                             [mt[0] + ut[0], mt[1]]])
        return dfdu
    
    def func_noise(self, var_sysd=1e-9):
        """Returns a function that returns a state-dependent Gaussian noise."""
        def fn(mt):
            gaussian_noise = pomdp_py.Gaussian([0,0],
                                               [[var_sysd, 0],
                                                [0, var_sysd]])
            return gaussian_noise
        return fn



"""Largely based on MosViz, except this is not an OO-POMDP"""
import pygame
import cv2
import math
import numpy as np
import random
import pomdp_py.utils as util
from pomdp_problems.tag.env.env import *
from pomdp_problems.tag.domain.observation import *
from pomdp_problems.tag.domain.action import *
from pomdp_problems.tag.domain.state import *
from pomdp_problems.tag.example_worlds import *
from pomdp_problems.tag.models.observation_model import *

#### Visualization through pygame ####
class TagViz:

    def __init__(self, env, res=30, fps=30, controllable=False, observation_model=None):
        self._env = env

        self._res = res
        self._img = self._make_gridworld_image(res)
        self._last_observation = None
        self._last_action = None
        self._last_belief = None
        self._observation_model = observation_model

        self._controllable = controllable
        self._running = False
        self._fps = fps
        self._playtime = 0.0

        self._target_color = (200, 0, 50)

    def _make_gridworld_image(self, r):
        # Preparing 2d array
        w, l = self._env.width, self._env.length
        arr2d = np.full((self._env.width,
                         self._env.length), 0)  # free grids
        # Creating image
        img = np.full((w*r,l*r,3), 255, dtype=np.int32)
        for x in range(w):
            for y in range(l):
                if (x,y) not in self._env.grid_map.obstacle_poses:
                    arr2d[x,y] == 0    # free
                    cv2.rectangle(img, (y*r, x*r), (y*r+r, x*r+r),
                                  (255, 255, 255), -1)
                else:
                    arr2d[x,y] == 1  # obstacle
                    cv2.rectangle(img, (y*r, x*r), (y*r+r, x*r+r),
                                  (40, 31, 3), -1)
                cv2.rectangle(img, (y*r, x*r), (y*r+r, x*r+r),
                              (0, 0, 0), 1, 8)
        return img

    @property
    def img_width(self):
        return self._img.shape[0]

    @property
    def img_height(self):
        return self._img.shape[1]

    @property
    def last_observation(self):
        return self._last_observation

    def update(self, action, observation, belief):
        """
        Update the visualization after there is new real action and observation
        and updated belief.
        """
        self._last_action = action
        self._last_observation = observation
        self._last_belief = belief

    @staticmethod
    def draw_robot(img, x, y, th, size, color=(255,12,12)):
        radius = int(round(size / 2))
        cv2.circle(img, (y+radius, x+radius), radius, color, thickness=6)
        # endpoint = (y+radius + int(round(radius*math.sin(th))),
        #             x+radius + int(round(radius*math.cos(th))))
        # cv2.line(img, (y+radius,x+radius), endpoint, color, 2)

    @staticmethod
    def draw_observation(img, z, rx, ry, rth, r, size, color=(12,12,255)):
        assert type(z) == TagObservation, "%s != TagObservation" % (str(type(z)))
        radius = int(round(r / 2))
        if z.target_position is not None:
            lx, ly = z.target_position
            cv2.circle(img, (ly*r+radius,
                             lx*r+radius), size, color, thickness=-1)


    # TODO! Deprecated.
    @staticmethod
    def draw_belief(img, belief, r, size, target_color):
        """belief (OOBelief)"""
        radius = int(round(r / 2))

        circle_drawn = {}  # map from pose to number of times drawn

        hist = belief.get_histogram()
        color = target_color

        last_val = -1
        count = 0
        for state in reversed(sorted(hist, key=hist.get)):
            if last_val != -1:
                color = util.lighter(color, 1-hist[state]/last_val)
            if np.mean(np.array(color) / np.array([255, 255, 255])) < 0.999:
                tx, ty = state.target_position
                if (tx,ty) not in circle_drawn:
                    circle_drawn[(tx,ty)] = 0
                circle_drawn[(tx,ty)] += 1

                cv2.circle(img, (ty*r+radius,
                                 tx*r+radius), size//circle_drawn[(tx,ty)], color, thickness=-1)
                last_val = hist[state]

                count +=1
                if last_val <= 0:
                    break

    # PyGame interface functions
    def on_init(self):
        """pygame init"""
        pygame.init()  # calls pygame.font.init()
        # init main screen and background
        self._display_surf = pygame.display.set_mode((self.img_width,
                                                      self.img_height),
                                                     pygame.HWSURFACE)
        self._background = pygame.Surface(self._display_surf.get_size()).convert()
        self._clock = pygame.time.Clock()

        # Font
        self._myfont = pygame.font.SysFont('Comic Sans MS', 30)
        self._running = True

    def on_event(self, event):
        if event.type == pygame.QUIT:
            self._running = False
        # TODO! DEPRECATED!
        elif event.type == pygame.KEYDOWN:
            u = None  # control signal according to motion model
            action = None  # control input by user

            if event.key == pygame.K_LEFT:
                action = MoveWest2D
            elif event.key == pygame.K_RIGHT:
                action = MoveEast2D
            elif event.key == pygame.K_DOWN:
                action = MoveSouth2D
            elif event.key == pygame.K_UP:
                action = MoveNorth2D
            elif event.key == pygame.K_SPACE:
                action = TagAction()

            if action is None:
                return

            if self._controllable:
                reward = self._env.state_transition(action, execute=True)
                robot_pose = self._env.state.robot_position
                z = None
                if self._observation_model is not None:
                    z = self._observation_model.sample(self._env.state, action)
                    self._last_observation = z
                print("      state: %s" % str(self._env.state))
                print("     action: %s" % str(action.name))
                print("     observation: %s" % str(z))
                print("     reward: %s" % str(reward))
                print(" valid motions: %s" % str(self._env.grid_map.valid_motions(self._env.state.robot_position)))
                print("------------")
                if self._env.state.target_found:
                    self._running = False
            return action

    def on_loop(self):
        self._playtime += self._clock.tick(self._fps) / 1000.0

    def on_render(self):
        # self._display_surf.blit(self._background, (0, 0))
        self.render_env(self._display_surf)
        rx, ry = self._env.state.robot_position
        fps_text = "FPS: {0:.2f}".format(self._clock.get_fps())
        last_action = self._last_action
        last_action_str = "no_action" if last_action is None else str(last_action)
        pygame.display.set_caption("%s | Robot(%.2f,%.2f,%.2f) | %s | %s" %
                                   (last_action_str, rx, ry, 0,
                                    str(self._env.state.target_found),
                                    fps_text))
        pygame.display.flip()

    def on_cleanup(self):
        pygame.quit()

    def on_execute(self):
        if self.on_init() == False:
            self._running = False

        while( self._running ):
            for event in pygame.event.get():
                self.on_event(event)
            self.on_loop()
            self.on_render()
        self.on_cleanup()

    def render_env(self, display_surf):
        img = np.copy(self._img)
        r = self._res  # Not radius! It's resolution.

        # draw target
        tx, ty = self._env.state.target_position
        cv2.rectangle(img, (ty*r, tx*r), (ty*r+r, tx*r+r),
                      (255, 165, 0), -1)

        # draw robot
        rx, ry = self._env.state.robot_position
        r = self._res  # Not radius!
        # last_observation = self._last_observation.get(robot_id, None)
        # last_viz_observation = self._last_viz_observation.get(robot_id, None)
        # last_belief = self._last_belief.get(robot_id, None)
        if self._last_belief is not None:
            TagViz.draw_belief(img, self._last_belief, r, r//3, self._target_color)
        if self._last_observation is not None:
            TagViz.draw_observation(img, self._last_observation,
                                    rx, ry, 0, r, r//8, color=(20, 20, 180))

        TagViz.draw_robot(img, rx*r, ry*r, 0, r, color=(200, 12, 150))
        pygame.surfarray.blit_array(display_surf, img)

# TODO! DEPRECATED!
def unittest():
    worldmap, robot = world0
    env = TagEnvironment.from_str(worldmap)
    observation_model = TagObservationModel()
    viz = TagViz(env, controllable=True, observation_model=observation_model)
    viz.on_execute()

if __name__ == '__main__':
    unittest()



"""
We implement POMCP as described in the original paper
Monte-Carlo Planning in Large POMDPs
https://papers.nips.cc/paper/4031-monte-carlo-planning-in-large-pomdps

One thing to note is that, in this algorithm, belief
update happens as the simulation progresses. The new
belief is stored in the vnodes at the level after
executing the next action. These particles will
be reinvigorated if they are not enough.

However, it is possible to separate MCTS completely
from the belief update. This means the belief nodes
no longer keep track of particles, and belief update
and particle reinvogration happen for once after MCTS
is completed. I have previously implemented this version.
This version is also implemented in BasicPOMCP.jl
(https://github.com/JuliaPOMDP/BasicPOMCP.jl)
The two should be EQUIVALENT. In general, it doesn't
hurt to do the belief update during MCTS, a feature
of using particle representation.
"""

from pomdp_py.framework.basics cimport Action, Agent, POMDP, State, Observation,\
    ObservationModel, TransitionModel, GenerativeDistribution, PolicyModel
from pomdp_py.framework.planner cimport Planner
from pomdp_py.representations.distribution.particles cimport Particles
from pomdp_py.representations.belief.particles cimport particle_reinvigoration
from pomdp_py.algorithms.po_uct cimport VNode, RootVNode, QNode, POUCT, RandomRollout
import copy
import time
import random
import math


cdef class VNodeParticles(VNode):
    """POMCP's VNode maintains particle belief"""
    def __init__(self, num_visits, belief=Particles([])):
        self.num_visits = num_visits
        self.belief = belief
        self.children = {}  # a -> QNode
    def __str__(self):
        return "VNode(%.3f, %.3f, %d | %s)" % (self.num_visits, self.value, len(self.belief),
                                               str(self.children.keys()))
    def __repr__(self):
        return self.__str__()

cdef class RootVNodeParticles(RootVNode):
    def __init__(self, num_visits, history, belief=Particles([])):
        # vnodeobj = VNodeParticles(num_visits, value, belief=belief)
        RootVNode.__init__(self, num_visits, history)
        self.belief = belief
    @classmethod
    def from_vnode(cls, vnode, history):
        rootnode = RootVNodeParticles(vnode.num_visits, history, belief=vnode.belief)
        rootnode.children = vnode.children
        return rootnode

cdef class POMCP(POUCT):

    """POMCP is POUCT + particle belief representation.
    This POMCP version only works for problems
    with action space that can be enumerated."""

    def __init__(self,
                 max_depth=5, planning_time=-1., num_sims=-1,
                 discount_factor=0.9, exploration_const=math.sqrt(2),
                 num_visits_init=0, value_init=0,
                 rollout_policy=RandomRollout(), action_prior=None,
                 show_progress=False, pbar_update_interval=5):
        super().__init__(max_depth=max_depth,
                         planning_time=planning_time,
                         num_sims=num_sims,
                         discount_factor=discount_factor,
                         exploration_const=exploration_const,
                         num_visits_init=num_visits_init,
                         value_init=value_init,
                         rollout_policy=rollout_policy,
                         action_prior=action_prior,
                         show_progress=show_progress,
                         pbar_update_interval=pbar_update_interval)

    @property
    def update_agent_belief(self):
        """True if planner's update function also updates agent's
        belief."""
        return True

    def plan(self, agent):
        # Only works if the agent's belief is particles
        if not isinstance(agent.belief, Particles):
            raise TypeError("Agent's belief is not represented in particles.\n"\
                            "POMCP not usable. Please convert it to particles.")
        return POUCT.plan(self, agent)

    cpdef update(self, Agent agent, Action real_action, Observation real_observation,
                 state_transform_func=None):
        """
        Assume that the agent's history has been updated after taking real_action
        and receiving real_observation.

        `state_transform_func`: Used to add artificial transform to states during
            particle reinvigoration. Signature: s -> s_transformed
        """
        if not isinstance(agent.belief, Particles):
            raise TypeError("agent's belief is not represented in particles.\n"\
                            "POMCP not usable. Please convert it to particles.")
        if not hasattr(agent, "tree"):
            print("Warning: agent does not have tree. Have you planned yet?")
            return

        if agent.tree[real_action][real_observation] is None:
            # Never anticipated the real_observation. No reinvigoration can happen.
            raise ValueError("Particle deprivation.")
        # Update the tree; Reinvigorate the tree's belief and use it
        # as the updated belief for the agent.
        agent.tree = RootVNodeParticles.from_vnode(agent.tree[real_action][real_observation],
                                                   agent.history)
        tree_belief = agent.tree.belief
        agent.set_belief(particle_reinvigoration(tree_belief,
                                                 len(agent.init_belief.particles),
                                                 state_transform_func=state_transform_func))
        # If observation was never encountered in simulation, then tree will be None;
        # particle reinvigoration will occur.
        if agent.tree is not None:
            agent.tree.belief = copy.deepcopy(agent.belief)

    cpdef _simulate(POMCP self,
                    State state, tuple history, VNode root, QNode parent,
                    Observation observation, int depth):
        total_reward = POUCT._simulate(self, state, history, root, parent, observation, depth)
        if depth == 1 and root is not None:
            root.belief.add(state)  # belief update happens as simulation goes.
        return total_reward

    def _VNode(self, agent=None, root=False, **kwargs):
        """Returns a VNode with default values; The function naming makes it clear
        that this function is about creating a VNode object."""
        if root:
            # agent cannot be None.
            return RootVNodeParticles(self._num_visits_init,
                                      agent.history,
                                      belief=copy.deepcopy(agent.belief))
        else:
            if agent is None:
                return VNodeParticles(self._num_visits_init,
                                      belief=Particles([]))
            else:
                return VNodeParticles(self._num_visits_init,
                                      belief=copy.deepcopy(agent.belief))



"""
The agent can take motion action and a look action.
"""

# Reuses the actions in the multi object search domain
import pomdp_py
from pomdp_problems.multi_object_search.domain.action\
    import MotionAction, MoveForward, MoveBackward, MoveLeft, MoveRight, LookAction



import argparse

available_problems = [
    'tiger',
    'rocksample',
    'mos',
    'tag',
    'load_unload'
]

def parse_args():
    parser = argparse.ArgumentParser(description="pomdp_py CLI")
    parser.add_argument("-r", "--run", type=str,
                        help="run a pomdp under pomdp_problems."
                        "Available options: {}".format(available_problems))
    args = parser.parse_args()
    return parser, args

if __name__ == "__main__":
    parser, args = parse_args()
    if args.run:
        if args.run.lower() == "tiger":
            from pomdp_problems.tiger.tiger_problem import main
            main()

        elif args.run.lower() == "rocksample":
            from pomdp_problems.rocksample.rocksample_problem import main
            main()

        elif args.run.lower() == "mos":
            from pomdp_problems.multi_object_search.problem import unittest
            unittest()

        elif args.run.lower() == "tag":
            from pomdp_problems.tag.problem import main
            main()

        elif args.run.lower() == "load_unload":
            from pomdp_problems.load_unload.load_unload import main
            main()

        else:
            print("Unrecognized pomdp: {}".format(args.run))

    else:
        parser.print_help()



from pomdp_py.framework.basics cimport GenerativeDistribution
import random

cdef class WeightedParticles(GenerativeDistribution):
    """
    Represents a distribution :math:`\Pr(X)` with weighted particles, each is a
    tuple (value, weight). "value" means a value for the random variable X. If
    multiple values are present for the same value, will interpret the
    probability at X=x as the average of those weights.

    __init__(self, list particles, str approx_method="none", distance_func=None)

    Args:
       particles (list): List of (value, weight) tuples. The weight represents
           the likelihood that the value is drawn from the underlying distribution.
       approx_method (str): 'nearest' if when querying the probability
            of a value, and there is no matching particle for it, return
            the probability of the value closest to it. Assuming values
            are comparable; "none" if no approximation, return 0.
       distance_func: Used when approx_method is 'nearest'. Returns
           a number given two values in this particle set.
    """
    def __init__(self, list particles, str approx_method="none", distance_func=None):
        self._values = [value for value, _ in particles]
        self._weights = [weight for _, weight in particles]
        self._particles = particles

        self._hist = self.get_histogram()
        self._hist_valid = True

        self._approx_method = approx_method
        self._distance_func = distance_func

    @property
    def particles(self):
        return self._particles

    @property
    def values(self):
        return self._values

    @property
    def weights(self):
        return self._weights

    def add(self, particle):
        """add(self, particle)
        particle: (value, weight) tuple"""
        self._particles.append(particle)
        s, w = particle
        self._values.append(s)
        self._weights.append(w)
        self._hist_valid = False

    def __str__(self):
        return str(self.condense().particles)

    def __len__(self):
        return len(self._particles)

    def __getitem__(self, value):
        """Returns the probability of `value`; normalized"""
        if len(self.particles) == 0:
            raise ValueError("Particles is empty.")

        if not self._hist_valid:
            self._hist = self.get_histogram()
            self._hist_valid = True

        if value in self._hist:
            return self._hist[value]
        else:
            if self._approx_method == "none":
                return 0.0
            elif self._approx_method == "nearest":
                nearest_dist = float('inf')
                nearest = self._values[0]
                for s in self._values[1:]:
                    dist = self._distance_func(s, nearest)
                    if dist < nearest_dist:
                        nearest_dist = dist
                        nearest = s
                return self[nearest]
            else:
                raise ValueError("Cannot handle approx_method:",
                                 self._approx_method)

    def __setitem__(self, value, prob):
        """
        The particle belief does not support assigning an exact probability to a value.
        """
        raise NotImplementedError

    def random(self):
        """Samples a value based on the particles"""
        value = random.choices(self._values, weights=self._weights, k=1)[0]
        return value

    def mpe(self):
        if not self._hist_valid:
            self._hist = self.get_histogram()
            self._hist_valid = True
        return max(self._hist, key=self._hist.get)

    def __iter__(self):
        return iter(self._particles)

    cpdef dict get_histogram(self):
        """
        get_histogram(self)
        Returns a mapping from value to probability, normalized."""
        cdef dict hist = {}
        cdef dict counts = {}
        # first, sum the weights
        for s, w in self._particles:
            hist[s] = hist.get(s, 0) + w
            counts[s] = counts.get(s, 0) + 1
        # then, average the sums
        total_weights = 0.0
        for s in hist:
            hist[s] = hist[s] / counts[s]
            total_weights += hist[s]
        # finally, normalize
        for s in hist:
            hist[s] /= total_weights
        return hist

    @classmethod
    def from_histogram(cls, histogram):
        """Given a pomdp_py.Histogram return a particle representation of it,
        which is an approximation"""
        particles = []
        for v in histogram:
            particles.append((v, histogram[v]))
        return WeightedParticles(particles)

    def condense(self):
        """
        Returns a new set of weighted particles with unique values
        and weights aggregated (taken average).
        """
        return WeightedParticles.from_histogram(self.get_histogram())


cdef class Particles(WeightedParticles):
    """ Particles is a set of unweighted particles; This set of particles represent
    a distribution :math:`\Pr(X)`. Each particle takes on a specific value of :math:`X`.
    Inherits :py:mod:`~pomdp_py.representations.distribution.particles.WeightedParticles`.

    __init__(self, particles, **kwargs)

    Args:
        particles (list): List of values.
        kwargs: see __init__() of :py:mod:`~pomdp_py.representations.distribution.particles.WeightedParticles`.
    """
    def __init__(self, particles, **kwargs):
        super().__init__(list(zip(particles, [None]*len(particles))), **kwargs)

    def __iter__(self):
        return iter(self.particles)

    def add(self, particle):
        """add(self, particle)
        particle: just a value"""
        self._particles.append((particle, None))
        self._values.append(particle)
        self._weights.append(None)
        self._hist_valid = False

    @property
    def particles(self):
        """For unweighted particles, the particles are just values."""
        return self._values

    def get_abstraction(self, state_mapper):

        """get_abstraction(self, state_mapper)
        feeds all particles through a state abstraction function.
        Or generally, it could be any function.
        """
        particles = [state_mapper(s) for s in self.particles]
        return particles

    @classmethod
    def from_histogram(cls, histogram, num_particles=1000):
        """Given a pomdp_py.Histogram return a particle representation of it,
        which is an approximation"""
        particles = []
        for _ in range(num_particles):
            particles.append(histogram.random())
        return Particles(particles)

    cpdef dict get_histogram(self):
        cdef dict hist = {}
        for s in self.particles:
            hist[s] = hist.get(s, 0) + 1
        for s in hist:
            hist[s] = hist[s] / len(self.particles)
        return hist

    def random(self):
        """Samples a value based on the particles"""
        if len(self._particles) > 0:
            return random.choice(self._values)
        else:
            return None



from pomdp_problems.light_dark.env.env import LightDarkEnvironment
from pomdp_problems.light_dark.env.visual import LightDarkViz
from pomdp_problems.light_dark.domain.state import *
from pomdp_problems.light_dark.domain.action import *
from pomdp_problems.light_dark.domain.observation import *
from pomdp_problems.light_dark.models.transition_model import *
from pomdp_problems.light_dark.models.observation_model import *



from pomdp_problems.tag.domain.action import *
from pomdp_problems.tag.models.transition_model import TagTransitionModel
from pomdp_problems.multi_object_search.env.env import interpret

class GridMap:

    def __init__(self, width, length, obstacle_poses):
        self.width = width
        self.length = length
        # set of obstacle poses
        self.obstacle_poses = obstacle_poses

    def valid_pose(self, position):
        if not (position[0] >= 0 and position[0] < self.width\
                and position[1] >= 0 and position[1] < self.length):
            return False
        if position in self.obstacle_poses:
            return False
        return True

    def valid_motions(self, position,
                      all_motions=MOTION_ACTIONS):
        valid_motions = set({})
        for motion_action in all_motions:
            if TagTransitionModel.if_move_by(self, position, motion_action) == position:
                continue
            valid_motions.add(motion_action)
        return valid_motions

    @classmethod
    def from_str(cls, worldstr, **kwargs):
        dim, _, objects, obstacles, _ = interpret(worldstr)
        obstacle_poses = set({})
        for objid in objects:
            if objid in obstacles:
                obstacle_poses.add(objects[objid].pose)
        grid_map = GridMap(dim[0], dim[1], obstacle_poses)
        return grid_map

    def free_cells(self):
        cells = set({(x,y)
                     for x in range(self.width)
                     for y in range(self.length)
                     if (x,y) not in self.obstacle_poses})
        return cells
                



"""Implementation of B-LQR algorithm described in "Belief space planning
assuming maximum likelihood observations" :cite:`platt2010belief`
"""

import pomdp_py
import numpy as np
from scipy import optimize

class BLQR(pomdp_py.Planner):

    def __init__(self,
                 func_sysd, func_obs, jac_sysd, jac_obs, jac_sysd_u,
                 noise_obs, noise_sysd,
                 Qlarge, L, Q, R,
                 planning_horizon=15):
        """To initialize the planenr of B-LQR, one needs to supply parameters
        that describe the underlying linear gaussian system, and also
        matrices that describe the cost function. Note that math symbols
        (e.g. xt, ut) are vectors or matrices represented as np.arrays. 
        The B-LQR cost function (Eq.14 in the paper):

        :math:`J(b_{t:T},u_{t:T}) = \bar{m}_T^TQ_{large}\bar{m}_T+\bar{s}_T^T\Lambda\bar{s}_T + \sum_{k=t}^{T-1}\bar{m}_k^TQ\bar{m}_k+\bar{u}_t^TR\bar{u}_t`
                    
        Args:
            func_sysd (function): f: (xt, ut) -> xt+1
            func_obs (function): g: (xt) -> zt
            jac_sysd (function): dfdx: (mt, ut) -> At
            jac_obs (function): dgdx: (mt) -> Ct
            jac_sysd_u (function): dfdu (mt, ut) -> Bt
            noise_obs (function): (xt) -> pomdp_py.Gaussian. Potentially
                state-dependent observation noise
            noise_sysd (function): (xt) -> pomdp_py.Gaussian. Potentially
                state-dependent system dynamics noise
            Qlarge (np.array) matrix of shape :math:`d \times d` where :math:`d`
                is the dimensionality of the state vector.
            L (np.array) matrix of shape :math:`d \times d` where :math:`d`
                is the dimensionality of the state vector.
                (Same as :math:`\Lambda` in the equation)
            Q (np.array) matrix of shape :math:`d \times d` where :math:`d`
                is the dimensionality of the state vector.
            R (np.array) matrix of shape :math:`c \times c` where :math:`c`
                is the dimensionality of the control vector.
            planning_horizon (int): This is the :math:`T`, the planning horizon.

        """
        self._func_sysd = func_sysd
        self._func_obs = func_obs
        self._jac_sysd = jac_sysd
        self._jac_obs = jac_obs
        self._jac_sysd = jac_sysd_u
        self._noise_obs = noise_obs
        self._noise_sysd = noise_sysd        
        self._Qlarge = Qlarge
        self._L = L  # Lambda
        self._Q = Q
        self._R = R
        self._planning_horizon = planning_horizon
        self._dim_state = self._Q.shape[0]
        self._dim_control = self._R.shape[0]

        
    def ekf_update_mlo(self, bt, ut):
        """
        Performs the ekf belief update assuming maximum likelihood observation.
        This follows equations 12, 13 in the paper. It's the function :math:`F`.

        Args:
            bt (tuple): a belief point bt = (mt, Cov_t) representing the belief state.
                where mt is np.array of shape (d,) and Cov_t is np.array of shape (d,d).
                The cost function equation needs to turn Cov_t into a long vector consist
                of the columns of Cov_t stiched together.
            ut (np.array): control vector
            noise_t (pomdp_py.Gaussian): The Gaussian noise with "possibly state-dependent"
                covariance matrix Wt.
            noise_sysd (np.array): A noise term (e.g. Gaussian noise) added to the
                system dynamics (:math:`v` in Eq.12). This array should have the sam
                dimensionality as mt.
            noise_obs_cov (np.array): The covariance matrix of the Gaussian noise of the
                observation function. This corresponds to Wt in equation 13.
        """
        # TODO: FIX
        mt, Cov_t = bt        
        At = self._jac_sysd(mt, ut)
        Ct = self._jac_obs(mt)  # based on maximum likelihood observation
        Wt = self._noise_obs(mt).cov
        Gat = np.dot(np.dot(At, Cov_t), At.transpose())  # Gamma_t = At*Cov_t*At^T
        CGC_W_inv = np.linalg.inv(np.dot(np.dot(Ct, Gat), Ct.transpose()) + Wt)
        Cov_next = Gat - np.dot(np.dot(np.dot(np.dot(Gat, Ct.transpose()), CGC_W_inv), Ct), Gat)
        m_next = self._func_sysd(mt, ut) + self._noise_sysd(mt).random()
        return (m_next, Cov_next)

    def _b_u_seg(self, x, i):
        """Returns the elements of x that corresponds to the belief and controls,
        as a tuple (m_i, Covi, u_i)"""
        d = self._dim_state
        c = self._dim_control
        start = i*(d + (d*d) + c)
        end = (i+1)*(d + (d*d) + c)
        m_i_end = i*(d + (d*d) + c) + d
        Covi_end = i*(d + (d*d) + c) + d + d*d
        u_end = i*(d + (d*d) + c) + d + d*d + c
        return x[start:m_i_end], x[m_i_end:Covi_end], x[Covi_end:u_end]
    
    def _control_max_constraint(self, x, i, max_val):
        _, _, u_i = self._b_u_seg(x, i)
        return max_val - u_i

    def _control_min_constraint(self, x, i, min_val):
        _, _, u_i = self._b_u_seg(x, i)
        return u_i - min_val

    def _mean_final_constraint(self, x, m_des, num_segments):
        m_k, _, _ = self._b_u_seg(x, num_segments-1)
        return m_k - m_des
    
    def _mean_start_constraint(self, x, m_0):
        m_k, _, _ = self._b_u_seg(x, 0)
        return m_k - m_0

    def _belief_constraint(self, x, i, num_segments):
        m_i, s_i, u_i = self._b_u_seg(x, i)
        Cov_i = s_i.reshape(self._dim_state, self._dim_state).transpose()
        m_ip1, Cov_ip1 = self.integrate_belief_segment((m_i, Cov_i), u_i, num_segments)
        s_ip1 = Cov_ip1.transpose().reshape(-1,)
        b_i_vec = np.append(m_i, s_i)
        b_ip1_vec = np.append(m_ip1, s_ip1)
        return b_i_vec - b_ip1_vec

    def integrate_belief_segment(self, b_i, u_i, num_segments):
        """This is to represent equation 18.
             
        phi(b_i', u_i') = F(b_i', u_i') +   sum    F(b_{t+1}, u_i) - F(b_t, u_i)
                                          t in seg

        This essentially returns b_{i+1}'
        """
        m_i, Cov_i = b_i
        b_seg = [self.ekf_update_mlo((m_i, Cov_i), u_i)]
        for t in range(1, self._planning_horizon // num_segments):
            b_seg.append(self.ekf_update_mlo(b_seg[t-1], u_i))
        sum_mean = b_seg[0][0]
        sum_Cov = b_seg[0][1]
        for t in range(len(b_seg)):
            if t + 1 < len(b_seg) - 1:
                sum_mean += (b_seg[t+1][0] - b_seg[t][0])
                sum_Cov += (b_seg[t+1][1] - b_seg[t][1])
        return sum_mean, sum_Cov

    
    def _opt_cost_func_seg(self, x, b_des, u_des, num_segments):
        """This will be used as part of the objective function for scipy
        optimizer. Therefore we only take in one input vector, x.  We require
        the structure of x to be:

        [ m1 | Cov 1 | u1 ] ... [ m_i | Cov i | u_i ]

        Use the _b_u_seg function to obtain the individual parts.
        """
        bu_traj = []
        for i in range(num_segments):
            m_i, s_i, u_i = self._b_u_seg(x, i)
            Cov_i = s_i.reshape(self._dim_state, self._dim_state).transpose()
            bu_traj.append(((m_i, Cov_i), u_i))
        return self.segmented_cost_function(bu_traj, b_des, u_des, num_segments)

    def segmented_cost_function(self, bu_traj, b_des, u_des, num_segments):
        """The cost function in eq 17. 

        Args:
            b_des (tuple): The desired belief (mT, CovT). This is needed to compute the cost function.
            u_des (list): A list of desired controls at the beginning of each segment.
                If this information is not available, pass in an empty list.
        """
        if len(u_des) > 0 and len(u_des) != num_segments:
            raise ValueError("The list of desired controls, if provided, must have one control"\
                             "per segment")

        b_T, u_T = bu_traj[-1]
        m_T, Cov_T = b_T
        s_T = Cov_T.transpose().reshape(-1,)
        
        sLs = np.dot(np.dot(s_T.transpose(), self._L), s_T)
        Summation = 0
        for i in range(num_segments):
            b_i, u_i = bu_traj[i]
            m_i, _ = b_i
            m_i_diff = m_i - b_des[0]
            if len(u_des) > 0:
                u_i_des = u_des[i]
            else:
                u_i_des = np.zeros(self._dim_control)
            u_i_diff = u_i - u_i_des
            Summation += np.dot(np.dot(m_i_diff.transpose(), self._Q), m_i_diff)\
                + np.dot(np.dot(u_i_diff.transpose(), self._R), u_i_diff)
        return sLs + Summation
    
    def create_plan(self, b_0, b_des, u_init, num_segments=5, control_bounds=None,
                    opt_options={}, opt_callback=None):
        """Solves the SQP problem using direct transcription, and produce belief points
        and controls at segments.
        Reference: https://docs.scipy.org/doc/scipy/reference/tutorial/optimize.html"""
        # build initial guess from initial belief and the given trajectory.
        x_0 = []
        b_i = b_0
        for i in range(num_segments):
            m_i, Cov_i = b_i
            s_i = Cov_i.transpose().reshape(-1,)
            u_i = u_init[i]
            x_0.extend(m_i)
            x_0.extend(s_i)
            x_0.extend(u_i)
            b_i = self.integrate_belief_segment(b_i, u_i, num_segments)

        # constraints        
        constraints = []
        ## initial belief constraint
        cons_start = {'type': 'eq',
                      'fun': self._mean_start_constraint,
                      'args': [b_0[0]]}
        constraints.append(cons_start)
        ## belief dynamics constraints and control bounds
        for i in range(num_segments):
            if i + 1 < num_segments-1:
                cons_belief = {'type': 'eq',
                               'fun': self._belief_constraint,
                               'args': [i, num_segments]}
                constraints.append(cons_belief)

                # control bounds
                if control_bounds is not None:
                    cons_control_min = {'type': 'ineq',
                                        'fun': self._control_min_constraint,
                                        'args': [i, control_bounds[0]]}                
                    cons_control_max = {'type': 'ineq',
                                        'fun': self._control_max_constraint,
                                        'args': [i, control_bounds[1]]}
                    constraints.append(cons_control_min)
                    constraints.append(cons_control_max)

        # final belief constraint
        cons_final = {'type': 'eq',
                      'fun': self._mean_final_constraint,
                      'args': [b_des[0], num_segments]}
        constraints.append(cons_final)

        opt_res = optimize.minimize(self._opt_cost_func_seg, x_0,
                                    method="SLSQP",
                                    args=(b_des, u_init, num_segments),
                                    constraints=constraints,
                                    options=opt_options,
                                    callback=opt_callback)
        return opt_res

    def interpret_sqp_plan(self, opt_res, num_segments):
        x_res = opt_res.x
        plan = []
        for i in range(num_segments):
            m_i, s_i, u_i = self._b_u_seg(x_res, i)
            Cov_i = s_i.reshape(self._dim_state, self._dim_state).transpose()
            plan.append((m_i, Cov_i, u_i))
        return plan



import pomdp_py
from pomdp_problems.tag.domain.action import *

class TagRewardModel(pomdp_py.RewardModel):

    def __init__(self, small=1, big=10):
        self.small = small
        self.big = big

    def probability(self, reward, state, action, next_state, normalized=False, **kwargs):
        if reward == self._reward_func(state, action):
            return 1.0
        else:
            return 0.0
        
    def sample(self, state, action, next_state):
        # deterministic
        return self._reward_func(state, action, next_state)

    def _reward_func(self, state, action, next_state):
        if isinstance(action, MotionAction):
            return -self.small
        else:
            # Tag action
            assert isinstance(action, TagAction)
            if next_state.target_position == next_state.robot_position:
                if next_state.target_found:
                    return self.big
            return -self.big



"""The Tag problem. Implemented according to the paper `Anytime Point-Based
Approximations for Large POMDPs <https://arxiv.org/pdf/1110.0027.pdf>`_.

Action space: The agent can take motion action and a tag action.
"""

# Reuses the actions in the multi object search domain
import pomdp_py
from pomdp_problems.multi_object_search.domain.action\
    import Action, MotionAction, MoveEast2D, MoveWest2D, MoveSouth2D, MoveNorth2D

MOTION_ACTIONS = {MoveEast2D, MoveWest2D, MoveSouth2D, MoveNorth2D}

class TagAction(Action):
    def __init__(self):
        super().__init__("tag")    



"""Utilities for dealing with colors"""

import numpy as np
import random

# colors
def lighter(color, percent):
    '''assumes color is rgb between (0, 0, 0) and (255, 255, 255)'''
    color = np.array(color)
    white = np.array([255, 255, 255])
    vector = white-color
    return color + vector * percent

def rgb_to_hex(rgb):
    r,g,b = rgb
    return '#%02x%02x%02x' % (int(r), int(g), int(b))

def hex_to_rgb(hx):
    """hx is a string, begins with #. ASSUME len(hx)=7."""
    if len(hx) != 7:
        raise ValueError("Hex must be #------")
    hx = hx[1:]  # omit the '#'
    r = int('0x'+hx[:2], 16)
    g = int('0x'+hx[2:4], 16)
    b = int('0x'+hx[4:6], 16)
    return (r,g,b)

def inverse_color_rgb(rgb):
    r,g,b = rgb
    return (255-r, 255-g, 255-b)

def inverse_color_hex(hx):
    """hx is a string, begins with #. ASSUME len(hx)=7."""
    return inverse_color_rgb(hex_to_rgb(hx))

def random_unique_color(colors, ctype=1):
    """
    ctype=1: completely random
    ctype=2: red random
    ctype=3: blue random
    ctype=4: green random
    ctype=5: yellow random
    """
    if ctype == 1:
        color = "#%06x" % random.randint(0x444444, 0x999999)
        while color in colors:
            color = "#%06x" % random.randint(0x444444, 0x999999)
    elif ctype == 2:
        color = "#%02x0000" % random.randint(0xAA, 0xFF)
        while color in colors:
            color = "#%02x0000" % random.randint(0xAA, 0xFF)
    elif ctype == 4:  # green
        color = "#00%02x00" % random.randint(0xAA, 0xFF)
        while color in colors:
            color = "#00%02x00" % random.randint(0xAA, 0xFF)
    elif ctype == 3:  # blue
        color = "#0000%02x" % random.randint(0xAA, 0xFF)
        while color in colors:
            color = "#0000%02x" % random.randint(0xAA, 0xFF)
    elif ctype == 5:  # yellow
        h = random.randint(0xAA, 0xFF)
        color = "#%02x%02x00" % (h, h)
        while color in colors:
            h = random.randint(0xAA, 0xFF)
            color = "#%02x%02x00" % (h, h)
    else:
        raise ValueError("Unrecognized color type %s" % (str(ctype)))
    return color



"""Defines the State for the maze domain, which is the position of the robot and its orientation.
"""
import pomdp_py
import numpy as np

class State(pomdp_py.State):
    """The state of the problem is just the robot position"""
    def __init__(self, positition, orientation):
        """
        Initializes a state in light dark domain.

        Args:
            position (tuple): position of the robot.
        """
        if len(position) != 2:
            raise ValueError("State position must be a vector of length 2")
        self.position = positition
        self.orientation = orientation

    def __hash__(self):
        return hash(self.position, self.orientation)
    
    def __eq__(self, other):
        if isinstance(other, State):
            return self.position == other.position
        else:
            return False
        
    def __str__(self):
        return self.__repr__()
    
    def __repr__(self):
        return "State(%s)" % (str(self.position, self.orientation))



from pomdp_py.framework.planner cimport Planner
from pomdp_py.framework.basics cimport Agent, PolicyModel, Action, State, Observation

cdef class TreeNode:
    cdef public dict children
    cdef public int num_visits
    cdef public float value

cdef class QNode(TreeNode):
    pass

cdef class VNode(TreeNode):
    cpdef argmax(VNode self)

cdef class RootVNode(VNode):
    cdef public tuple history

cdef class POUCT(Planner):
    cdef int _max_depth
    cdef float _planning_time
    cdef int _num_sims
    cdef int _num_visits_init
    cdef float _value_init
    cdef float _discount_factor
    cdef float _exploration_const
    cdef ActionPrior _action_prior
    cdef RolloutPolicy _rollout_policy
    cdef Agent _agent
    cdef int _last_num_sims
    cdef float _last_planning_time
    cdef bint _show_progress
    cdef int _pbar_update_interval

    cpdef _search(self)
    cpdef _simulate(POUCT self,
                    State state, tuple history, VNode root, QNode parent,
                    Observation observation, int depth)

    cpdef _expand_vnode(self, VNode vnode, tuple history, State state=*)
    cpdef _rollout(self, State state, tuple history, VNode root, int depth)
    cpdef Action _ucb(self, VNode root)
    cpdef tuple _sample_generative_model(self, State state, Action action)

cdef class RolloutPolicy(PolicyModel):
    cpdef Action rollout(self, State state, tuple history)

cdef class RandomRollout(RolloutPolicy):
    pass

cdef class ActionPrior:
    cpdef get_preferred_actions(ActionPrior self, State state, tuple history)



"""The Tag problem. Implemented according to the paper `Anytime Point-Based
Approximations for Large POMDPs <https://arxiv.org/pdf/1110.0027.pdf>`_.

Observation space: the agent observes the target's location when the agent and
    the target are in the same cell.
"""
import pomdp_py

class TagObservation(pomdp_py.Observation):
    def __init__(self, target_position):
        self.target_position = target_position

    def __hash__(self):
        return hash(self.target_position)

    def __eq__(self, other):
        if not isinstance(other, TagObservation):
            return False
        else:
            return self.target_position == other.target_position

    def __str__(self):
        return 'Observation(%s)' % (str(self.target_position))




"""Sensor model (for example, laser scanner)"""

import math
import numpy as np
from pomdp_problems.multi_object_search.domain.action import *
from pomdp_problems.multi_object_search.domain.observation import *

# Note that the occlusion of an object is implemented based on
# whether a beam will hit an obstacle or some other object before
# that object. Because the world is discretized, this leads to
# some strange pattern of the field of view. But what's for sure
# is that, when occlusion is enabled, the sensor will definitely
# not receive observation for some regions in the field of view
# making it a more challenging situation to deal with.

# Utility functions
def euclidean_dist(p1, p2):
    return math.sqrt(sum([(a - b)** 2 for a, b in zip(p1, p2)]))

def to_rad(deg):
    return deg * math.pi / 180.0

def in_range(val, rang):
    # Returns True if val is in range (a,b); Inclusive.
    return val >= rang[0] and val <= rang[1]

#### Sensors ####
class Sensor:
    LASER = "laser"
    PROXIMITY = "proximity"
    def observe(self, robot_pose, env_state):
        """
        Returns an Observation with this sensor model.
        """
        raise NotImplementedError

    def within_range(self, robot_pose, point):
        """Returns true if the point is within range of the sensor; but the point might not
        actually be visible due to occlusion or "gap" between beams"""
        raise ValueError

    @property
    def sensing_region_size(self):
        return self._sensing_region_size

    @property
    def robot_id(self):
        # id of the robot equipped with this sensor
        return self._robot_id

class Laser2DSensor:
    """Fan shaped 2D laser sensor"""

    def __init__(self, robot_id,
                 fov=90, min_range=1, max_range=5,
                 angle_increment=5,
                 occlusion_enabled=False):
        """
        fov (float): angle between the start and end beams of one scan (degree).
        min_range (int or float)
        max_range (int or float)
        angle_increment (float): angular distance between measurements (rad).
        """
        self.robot_id = robot_id
        self.fov = to_rad(fov)  # convert to radian
        self.min_range = min_range
        self.max_range = max_range
        self.angle_increment = to_rad(angle_increment)
        self._occlusion_enabled = occlusion_enabled

        # determines the range of angles;
        # For example, the fov=pi, means the range scanner scans 180 degrees
        # in front of the robot. By our angle convention, 180 degrees maps to [0,90] and [270, 360]."""
        self._fov_left = (0, self.fov / 2)
        self._fov_right = (2*math.pi - self.fov/2, 2*math.pi)

        # beams that are actually within the fov (set of angles)
        self._beams = {round(th, 2)
                       for th in np.linspace(self._fov_left[0],
                                             self._fov_left[1],
                                             int(round((self._fov_left[1] - self._fov_left[0]) / self.angle_increment)))}\
                    | {round(th, 2)
                       for th in np.linspace(self._fov_right[0],
                                             self._fov_right[1],
                                             int(round((self._fov_right[1] - self._fov_right[0]) / self.angle_increment)))}
        # The size of the sensing region here is the area covered by the fan
        self._sensing_region_size = self.fov / (2*math.pi) * math.pi * (max_range - min_range)**2

    def in_field_of_view(th, view_angles):
        """Determines if the beame at angle `th` is in a field of view of size `view_angles`.
        For example, the view_angles=180, means the range scanner scans 180 degrees
        in front of the robot. By our angle convention, 180 degrees maps to [0,90] and [270, 360]."""
        fov_right = (0, view_angles / 2)
        fov_left = (2*math.pi - view_angles/2, 2*math.pi)

    def within_range(self, robot_pose, point):
        """Returns true if the point is within range of the sensor; but the point might not
        actually be visible due to occlusion or "gap" between beams"""
        dist, bearing = self.shoot_beam(robot_pose, point)
        if not in_range(dist, (self.min_range, self.max_range)):
            return False
        if (not in_range(bearing, self._fov_left))\
           and (not in_range(bearing, self._fov_right)):
            return False
        return True

    def shoot_beam(self, robot_pose, point):
        """Shoots a beam from robot_pose at point. Returns the distance and bearing
        of the beame (i.e. the length and orientation of the beame)"""
        rx, ry, rth = robot_pose
        dist = euclidean_dist(point, (rx,ry))
        bearing = (math.atan2(point[1] - ry, point[0] - rx) - rth) % (2*math.pi)  # bearing (i.e. orientation)
        return (dist, bearing)

    def valid_beam(self, dist, bearing):
        """Returns true beam length (i.e. `dist`) is within range and its angle
        `bearing` is valid, that is, it is within the fov range and in
        accordance with the angle increment."""
        return dist >= self.min_range and dist <= self.max_range\
            and round(bearing, 2) in self._beams

    def _build_beam_map(self, beam, point, beam_map={}):
        """beam_map (dict): Maps from bearing to (dist, point)"""
        dist, bearing = beam
        valid = self.valid_beam(dist, bearing)
        if not valid:
            return
        bearing_key = round(bearing,2)
        if bearing_key in beam_map:
            # There's an object covered by this beame already.
            # see if this beame is closer
            if dist < beam_map[bearing_key][0]:
                # point is closer; Update beam map
                print("HEY")
                beam_map[bearing_key] = (dist, point)
            else:
                # point is farther than current hit
                pass
        else:
            beam_map[bearing_key] = (dist, point)

    def observe(self, robot_pose, env_state):
        """
        Returns a MosObservation with this sensor model.
        """
        rx, ry, rth = robot_pose

        # Check every object
        objposes = {}
        beam_map = {}
        for objid in env_state.object_states:
            objposes[objid] = ObjectObservation.NULL
            object_pose = env_state.object_states[objid]["pose"]
            beam = self.shoot_beam(robot_pose, object_pose)

            if not self._occlusion_enabled:
                if self.valid_beam(*beam):
                    d, bearing = beam  # distance, bearing
                    lx = rx + int(round(d * math.cos(rth + bearing)))
                    ly = ry + int(round(d * math.sin(rth + bearing)))
                    objposes[objid] = (lx, ly)
            else:
                self._build_beam_map(beam, object_pose, beam_map=beam_map)

        if self._occlusion_enabled:
            # The observed objects are in the beam_map
            for bearing_key in beam_map:
                d, objid = beam_map[bearing_key]
                lx = rx + int(round(d * math.cos(rth + bearing_key)))
                ly = ry + int(round(d * math.sin(rth + bearing_key)))
                objposes[objid] = (lx, ly)

        return MosOOObservation(objposes)

    @property
    def sensing_region_size(self):
        return self._sensing_region_size


class ProximitySensor(Laser2DSensor):
    """This is a simple sensor; Observes a region centered
    at the robot."""
    def __init__(self, robot_id,
                 radius=5,
                 occlusion_enabled=False):
        """
        radius (int or float) radius of the sensing region.
        """
        self.robot_id = robot_id
        self.radius = radius
        self._occlusion_enabled = occlusion_enabled

        # This is in fact just a specific kind of Laser2DSensor
        # that has a 360 field of view, min_range = 0.1 and
        # max_range = radius
        if occlusion_enabled:
            angle_increment = 5
        else:
            angle_increment = 0.25
        super().__init__(robot_id,
                         fov=360,
                         min_range=0.1,
                         max_range=radius,
                         angle_increment=angle_increment,
                         occlusion_enabled=occlusion_enabled)



# """
# Defines the belief representation.

# Origin: Belief space planning assuming maximum likelihood observations

# Quote from the paper:

#     Belief state was modeled as an isotropic Gaussian pdf
#     over the state space: :math:`b = (m,s) \in \mathbb{R}^2\times\mathbb{R}^{+}`.

# Technical definition:

# * An **isotropic gaussian** is one where the covariance matrix is
#   represented by the simplified matrix `source <https://math.stackexchange.com/questions/1991961/gaussian-distribution-is-isotropic>`_

# This is an example where you implement a problem-specific
# belief representation, instead of directly using an existing
# one provided by the pomdp_py; The interfaces of pomdp_py does
# support this kind of extension. In this example, the belief
# distribution is a variant of pomdp_py.Gaussian.

# The pomdp_py library does not (currently) include a BeliefState interface,
# since the belief distribution itself is just a Distribution and adding
# such interface makes the library unnecessarily complicated. The existing
# interfaces are fully able to support belief space planning as well.
# """
# import pomdp_py

# class IsotropicGaussian2d(pomdp_py.Gaussian):

#     def __init__(self, mean, sigma):
#         """initialize isotropic Gaussian distribution.
        
#         Args:
#             mean (list): 2d vector of robot position
#             sigma (float): sigma defining the standard deviation of the Gaussian.
#         """
#         if len(mean) != 2:
#             raise ValueError("Mean vector of belief must be of length 2")
#         self._sigma = sigma
#         super().__init__(mean,
#                          [[sigma**2, 0,
#                            0, sigma**2]])
        
#     @property
#     def sigma(self):
#         return self._sigma

    
# class BeliefState(pomdp_py.State):
#     """Belief state;

#     **Note** that the paper makes use of belief-space reward function, R(b,a),
#     which treats a belief distribution as a state, we shall define a
#     BeliefState class that contains an underlying Distribution.  Also, we
#     define belief space dynamics models; See light_dark.models.belief_space.
#     """
#     def __init__(self, belief):
#         if type(belief) != IsotropicGaussian2d:
#             raise TypeError("belief state's belief distribution must be LightDarkBelief")

#         # The belief is LightDarkBelief, which is a Gaussian.
#         self._belief = belief

#         # For hashing, use the mean
#         self._hashcode = hash(tuple(self._belief.mean))

#     @property
#     def belief(self):
#         return self._belief
        
#     def __hash__(self):
#         return self._hashcode
    
#     def __eq__(self, other):
#         if isinstance(other, BeliefState):
#             return self._belief == other.belief
#         else:
#             return False
        
#     def __str__(self):
#         return self.__repr__()
    
#     def __repr__(self):
#         return "BeliefState(%s)" % (str(self._belief.mean, self._belief.sigma))
        



"""The Tag problem. Implemented according to the paper `Anytime Point-Based
Approximations for Large POMDPs <https://arxiv.org/pdf/1110.0027.pdf>`_.

State space: state of the robot (x,y), state of the person (x,y), person found.

"""
import pomdp_py

class TagState(pomdp_py.State):
    
    def __init__(self, robot_position, target_position, target_found):
        """
        robot_position (tuple): x,y location of the robot.
        target_position (tuple): x,y location of the target.
        target_found (bool): True if the target is found.
        """
        self.robot_position = robot_position
        self.target_position = target_position
        self.target_found = target_found

    def __hash__(self):
        return hash((self.robot_position, self.target_position, self.target_found))

    def __eq__(self, other):
        if not isinstance(other, TagState):
            return False
        else:
            return self.robot_position == other.robot_position\
                and self.target_position == other.target_position\
                and self.target_found == other.target_found

    def __str__(self):
        return 'State(%s, %s | %s)' % (str(self.robot_position),
                                       str(self.target_position),
                                       str(self.target_found))

    def __repr__(self):
        return str(self)



"""Utility functions for Cython code."""

import hashlib

cpdef det_dict_hash(dct, keep=9):
    """deterministic hash of a dictionary."""
    content = str(list(sorted(dct.items()))).encode()
    hashcode = int(str(int(hashlib.sha1(content).hexdigest(), 16))[:keep])
    return hashcode



from pomdp_py.framework.basics cimport GenerativeDistribution

cdef class WeightedParticles(GenerativeDistribution):
    cdef list _particles
    cdef list _values
    cdef list _weights
    cdef str _approx_method
    cdef object _distance_func
    cdef dict _hist
    cdef bint _hist_valid

    cpdef dict get_histogram(self)

cdef class Particles(WeightedParticles):
    pass



from pomdp_py.representations.distribution.particles cimport Particles
from pomdp_py.framework.basics cimport State, Action, Observation, Agent,\
    TransitionModel, ObservationModel, BlackboxModel
import copy

"""To update particle belief, there are two possibilities.
Either, an algorithm such as POMCP is used which will update
the belief automatically when the planner gets updated; Or,
a standalone particle belief distribution is updated. In
the latter case, the belief update algorithm is as described
in :cite:`silver2010monte` but written explicitly instead
of done implicitly when the MCTS tree truncates as in POMCP.
"""

def abstraction_over_particles(particles, state_mapper):
    particles = [state_mapper(s) for s in particles]
    return particles

cpdef particle_reinvigoration(Particles particles,
                              int num_particles, state_transform_func=None):
    """Note that particles should contain states that have already made
    the transition as a result of the real action. Therefore, they simply
    form part of the reinvigorated particles. At least maintain `num_particles`
    number of particles. If already have more, then it's ok.
    """
    # If not enough particles, introduce artificial noise to existing particles (reinvigoration)
    cdef Particles new_particles = copy.deepcopy(particles)
    if len(new_particles) == 0:
        raise ValueError("Particle deprivation.")

    if len(new_particles) > num_particles:
        return new_particles
    
    print("Particle reinvigoration for %d particles" % (num_particles - len(new_particles)))
    cdef State next_state    
    while len(new_particles) < num_particles:
        # need to make a copy otherwise the transform affects states in 'particles'
        next_state = copy.deepcopy(particles.random())
        # Add artificial noise
        if state_transform_func is not None:
            next_state = state_transform_func(next_state)
        new_particles.add(next_state)
    return new_particles


cpdef update_particles_belief(Particles current_particles,
                              Action real_action, Observation real_observation=None,
                              ObservationModel observation_model=None,
                              TransitionModel transition_model=None,
                              BlackboxModel blackbox_model=None,
                              state_transform_func=None):
    """
    update_particles_belief(Particles current_particles,
                           Action real_action, Observation real_observation=None,
                           ObservationModel observation_model=None,
                           TransitionModel transition_model=None,
                           BlackboxModel blackbox_model=None,
                           state_transform_func=None)
    This is the second case (update particles belief explicitly); Either
    BlackboxModel is not None, or TransitionModel and ObservationModel are not
    None. Note that you DON'T need to call this function if you are using POMCP.
    |TODO: not tested|

    Args:
        state_transform_func (State->State) is used to add artificial noise to
            the reinvigorated particles.
    """
    cdef State particle, next_state
    cdef Observation observation
    cdef list filtered_particles = []
    for particle in current_particles.particles:
        # particle represents a state
        if blackbox_model is not None:
            # We're using a blackbox generator; (s',o,r) ~ G(s,a)
            result = blackbox_model.sample(particle, real_action)
            next_state = result[0]
            observation = result[1]
        else:
            # We're using explicit models
            next_state = transition_model.sample(particle, real_action)
            observation = observation_model.sample(next_state, real_action)
        # If observation matches real, then the next_state is accepted
        if observation == real_observation:
            filtered_particles.append(next_state)
    # Particle reinvigoration
    return particle_reinvigoration(Particles(filtered_particles), len(current_particles.particles),
                                   state_transform_func=state_transform_func)
    
